"""Tiferet Data Transfer Object Settings"""

# *** imports

# ** core
from typing import Any

# ** infra
from schematics.models import Model

# ** app
from ..events import RaiseError, a

# *** constants

# ** constant: default_module_path
DEFAULT_MODULE_PATH = 'tiferet.contexts.app'

# ** constant: default_class_name
DEFAULT_CLASS_NAME = 'AppInterfaceContext'

# *** classes

# ** class: aggregate
class Aggregate(Model):
    '''
    A data representation of an aggregate object.
    '''

    # * method: new
    @staticmethod
    def new(
        aggregate_type: type,
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> 'Aggregate':
        '''
        Initializes a new aggregate object.

        :param aggregate_type: The type of aggregate object to create.
        :type aggregate_type: type
        :param validate: True to validate the aggregate object.
        :type validate: bool
        :param strict: True to enforce strict mode for the aggregate object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new aggregate object.
        :rtype: Aggregate
        '''

        # Create a new aggregate object.
        aggregate_object: Aggregate = aggregate_type(dict(
            **kwargs
        ), strict=strict)

        # Validate if specified.
        if validate:
            aggregate_object.validate()

        # Return the new aggregate object.
        return aggregate_object

    # * method: set_attribute
    def set_attribute(self, attribute: str, value: Any) -> None:
        '''
        Update an attribute on the aggregate object.

        Raises an error if the attribute does not exist on the instance.

        :param attribute: The attribute name to update.
        :type attribute: str
        :param value: The new value.
        :type value: Any
        :return: None
        :rtype: None
        '''

        # Check if the attribute exists on the aggregate instance.
        # If the attribute does not exist, raise an error.
        if not hasattr(self, attribute):
            RaiseError.execute(
                error_code=a.const.INVALID_MODEL_ATTRIBUTE_ID,
                attribute=attribute,
            )

        # Apply the update to the attribute.
        setattr(self, attribute, value)

        # Perform final aggregate validation.
        self.validate()

# ** class: transfer_object
class TransferObject(Model):
    '''
    A data representation object.
    '''

    # ** method: map
    def map(self,
            type: Aggregate,
            role: str = 'to_model',
            validate: bool = True,
            **kwargs
            ) -> Aggregate:
        '''
        Maps the model data to a model object.

        :param type: The type of aggregate object to map to.
        :type type: type
        :param role: The role for the mapping.
        :type role: str
        :param validate: True to validate the aggregate object.
        :type validate: bool
        :param kwargs: Additional keyword arguments for mapping.
        :type kwargs: dict
        :return: A new aggregate object.
        :rtype: Aggregate
        '''

        # Get primitive of the model data and merge with the keyword arguments.
        # Give priority to the keyword arguments.
        data_object = self.to_primitive(role=role)
        for key, value in kwargs.items():
            data_object[key] = value

        # Map the data object to an aggregate object.
        # Attempt to create a new aggregate object with a custom factory method.
        # If the factory method does not exist, employ the standard method.
        try:
            aggregate_object = type.new(**data_object, strict=False)
        except Exception:
            aggregate_object = Aggregate.new(type, **data_object, strict=False)

        # Validate if specified.
        if validate:
            aggregate_object.validate()

        # Return the aggregate object.
        return aggregate_object

    # ** method: from_model
    @staticmethod
    def from_model(
        transfer_obj: 'TransferObject',
        aggregate: Aggregate,
        validate: bool = True,
        **kwargs
    ) -> 'TransferObject':
        '''
        Initializes a new data object from a model object.

        :param transfer_obj: The type of transfer object to map from.
        :type transfer_obj: type
        :param aggregate: The aggregate object to map from.
        :type aggregate: Aggregate
        :param validate: True to validate the data object.
        :type validate: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new transfer object.
        :rtype: TransferObject
        '''

        # Convert the aggregate object to a primitive dictionary and merge with the keyword arguments.
        # Give priority to the keyword arguments.
        aggregate_data = aggregate.to_primitive()
        for key, value in kwargs.items():
            aggregate_data[key] = value

        # Create a new transfer object.
        data_object = transfer_obj(dict(
            **aggregate_data
        ), strict=False)

        # Validate the data object if specified.
        if validate:
            data_object.validate()

        # Return the data object.
        return data_object

    # ** method: from_data
    @staticmethod
    def from_data(
        data: type,
        **kwargs
    ) -> 'TransferObject':
        '''
        Initializes a new transfer object from a dictionary.

        :param data: The type of transfer object to map from.
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new transfer object.
        :rtype: TransferObject
        '''

        # Create a new transfer object.
        return data(dict(**kwargs), strict=False)

    # ** method: allow
    @staticmethod
    def allow(*args) -> Any:
        '''
        Creates a whitelist transform for data mapping.

        :param args: Fields to allow in the transform.
        :type args: tuple
        :return: The whitelist transform.
        :rtype: Any
        '''

        # Create a whitelist transform.
        # Create a wholelist transform if no arguments are specified.
        from schematics.transforms import whitelist, wholelist
        if args:
            return whitelist(*args)
        return wholelist()

    # ** method: deny
    @staticmethod
    def deny(*args) -> Any:
        '''
        Creates a blacklist transform for data mapping.

        :param args: Fields to deny in the transform.
        :type args: tuple
        :return: The blacklist transform.
        :rtype: Any
        '''

        # Create a blacklist transform.
        from schematics.transforms import blacklist
        return blacklist(*args)



"""Tiferet Utils File"""

# *** imports

# ** core
from pathlib import Path
from typing import IO, Any, Optional

# ** app
from ..interfaces.file import FileService
from ..events import RaiseError, a

# *** utils

# ** util: file_loader
class FileLoader(FileService):
    '''
    Base utility for low-level file stream operations with validation and context support.
    Implements the FileService contract.
    '''

    # * attribute: path
    path: Path

    # * attribute: mode
    mode: str

    # * attribute: encoding
    encoding: Optional[str]

    # * attribute: newline
    newline: Optional[str]

    # * attribute: file
    file: Optional[IO[Any]]

    # * init
    def __init__(self,
            path: str | Path,
            mode: str = 'r',
            encoding: Optional[str] = None,
            newline: Optional[str] = None,
            **kwargs,
        ):
        '''
        Initialize FileLoader with path and open parameters.

        :param path: File system path (str or Path).
        :type path: str | Path
        :param mode: File open mode.
        :type mode: str
        :param encoding: Text encoding for text modes.
        :type encoding: Optional[str]
        :param newline: Newline handling mode.
        :type newline: Optional[str]
        :param kwargs: Additional parameters (ignored).
        :type kwargs: dict
        '''

        # Set the file path as a Path object.
        self.path = Path(path)

        # Set the file mode.
        self.mode = mode

        # Set the encoding for text modes.
        self.encoding = encoding

        # Set the newline handling mode.
        self.newline = newline

        # Initialize the file stream to None.
        self.file = None

    # * method: verify_file (static)
    @staticmethod
    def verify_file(path: Path, mode: str = 'r'):
        '''
        Verify that the file or parent directory exists as appropriate.

        For write/append/exclusive modes, verifies the parent directory exists.
        For read modes, verifies the file itself exists.

        :param path: The file path to verify.
        :type path: Path
        :param mode: The file open mode (used to determine verification strategy).
        :type mode: str
        '''

        # For write, append, or exclusive modes, verify the parent directory exists.
        if any(c in mode for c in ('w', 'a', 'x')):
            if not path.parent.exists():
                RaiseError.execute(
                    error_code=a.const.FILE_NOT_FOUND_ID,
                    path=str(path),
                )

        # For read modes, verify the file itself exists.
        else:
            if not path.exists():
                RaiseError.execute(
                    error_code=a.const.FILE_NOT_FOUND_ID,
                    path=str(path),
                )

    # * method: verify_mode
    def verify_mode(self):
        '''
        Validate the file mode string.

        :raises TiferetError: If the mode is not in the set of valid modes.
        '''

        # Define the set of valid file modes.
        valid = {
            'r', 'rb', 'w', 'wb', 'a', 'ab',
            'x', 'xb', 'r+', 'rb+', 'w+', 'wb+', 'a+', 'ab+',
        }

        # Raise an error if the mode is not valid.
        if self.mode not in valid:
            RaiseError.execute(
                error_code=a.const.INVALID_FILE_MODE_ID,
                mode=self.mode,
            )

    # * method: verify_encoding
    def verify_encoding(self):
        '''
        Ensure encoding is provided when required for text modes.

        :raises TiferetError: If encoding is None for a text (non-binary) mode.
        '''

        # Raise an error if encoding is missing for a text mode.
        if 'b' not in self.mode and self.encoding is None:
            RaiseError.execute(
                error_code=a.const.INVALID_ENCODING_ID,
                encoding=None,
            )

    # * method: open_file
    def open_file(self) -> IO[Any]:
        '''
        Open the file stream if not already open.

        :return: The opened file stream.
        :rtype: IO[Any]
        :raises TiferetError: If the file is already open, the path is invalid,
            the mode is invalid, or encoding is missing for text modes.
        '''

        # Raise an error if the file is already open.
        if self.file is not None:
            RaiseError.execute(
                error_code=a.const.FILE_ALREADY_OPEN_ID,
                path=str(self.path),
            )

        # Verify the file path exists as appropriate for the mode.
        self.verify_file(self.path, self.mode)

        # Validate the file mode.
        self.verify_mode()

        # Validate the encoding for text modes.
        self.verify_encoding()

        # Open the file with the specified parameters.
        self.file = open(
            self.path,
            mode=self.mode,
            encoding=self.encoding,
            newline=self.newline,
        )

        # Return the opened file stream.
        return self.file

    # * method: close_file
    def close_file(self) -> None:
        '''
        Close the open file stream if present.
        '''

        # Close the file if it is open and reset the attribute.
        if self.file is not None:
            self.file.close()
            self.file = None

    # * method: __enter__
    def __enter__(self) -> IO[Any]:
        '''
        Enter the runtime context and open the file stream.

        :return: The opened file stream.
        :rtype: IO[Any]
        '''

        # Open and return the file stream.
        return self.open_file()

    # * method: __exit__
    def __exit__(self, exc_type, exc_val, exc_tb):
        '''
        Exit the runtime context and close the file stream.

        :param exc_type: The exception type (if any).
        :param exc_val: The exception value (if any).
        :param exc_tb: The exception traceback (if any).
        :return: False to propagate exceptions.
        :rtype: bool
        '''

        # Close the file stream.
        self.close_file()

        # Do not suppress exceptions.
        return False



"""Tiferet Interfaces File"""

# *** imports

# ** core
from abc import abstractmethod
from typing import IO, Any

# ** app
from .settings import Service

# *** interfaces

# ** interface: file_service
class FileService(Service):
    '''
    Service contract for low-level file stream operations.
    '''

    # * method: open_file
    @abstractmethod
    def open_file(self, path: str, mode: str = 'r', encoding: str | None = None, **kwargs) -> IO[Any]:
        '''
        Open a file stream.

        :param path: File system path.
        :type path: str
        :param mode: File open mode (r, w, a, etc.).
        :type mode: str
        :param encoding: Text encoding (for text modes).
        :type encoding: str | None
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: Open file object.
        :rtype: IO[Any]
        '''
        raise NotImplementedError()

    # * method: close_file
    @abstractmethod
    def close_file(self, file: IO[Any]) -> None:
        '''
        Safely close an open file stream.

        :param file: The open file object to close.
        :type file: IO[Any]
        '''
        raise NotImplementedError()

    # * method: __enter__
    @abstractmethod
    def __enter__(self) -> 'FileService':
        '''
        Enter the context manager.

        :return: The file service instance.
        :rtype: FileService
        '''
        raise NotImplementedError()

    # * method: __exit__
    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        '''
        Exit the context manager.

        :param exc_type: The exception type, if any.
        :param exc_val: The exception value, if any.
        :param exc_tb: The exception traceback, if any.
        '''
        raise NotImplementedError()



"""Tiferet DI Mappers"""

# *** imports

# ** core
from typing import Dict, Any

# ** app
from ..domain import (
    FlaggedDependency,
    ServiceConfiguration,
    DomainObject,
    StringType,
    DictType,
    ModelType,
)
from .settings import (
    Aggregate,
    TransferObject,
)

# *** mappers

# ** mapper: flagged_dependency_aggregate
class FlaggedDependencyAggregate(FlaggedDependency, Aggregate):
    '''
    An aggregate representation of a flagged dependency.
    '''

    # * method: new
    @staticmethod
    def new(
        flagged_dependency_data: Dict[str, Any],
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> 'FlaggedDependencyAggregate':
        '''
        Initializes a new flagged dependency aggregate.

        :param flagged_dependency_data: The data to create the flagged dependency aggregate from.
        :type flagged_dependency_data: dict
        :param validate: True to validate the aggregate object.
        :type validate: bool
        :param strict: True to enforce strict mode for the aggregate object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new flagged dependency aggregate.
        :rtype: FlaggedDependencyAggregate
        '''

        # Create a new flagged dependency aggregate from the provided data.
        return Aggregate.new(
            FlaggedDependencyAggregate,
            validate=validate,
            strict=strict,
            **flagged_dependency_data,
            **kwargs
        )

    # * method: set_parameters
    def set_parameters(self, parameters: Dict[str, Any] | None = None) -> None:
        '''
        Update the parameters dictionary for this flagged dependency.

        :param parameters: New parameters, or None to clear all.
                           Keys with None values are removed.
        :type parameters: Dict[str, Any] | None
        :return: None
        :rtype: None
        '''

        # If parameters is None, clear all.
        if parameters is None:
            self.parameters = {}

        else:
            # Merge existing parameters with new ones (new values win).
            merged = dict(self.parameters or {})
            merged.update(parameters)

            # Filter out keys where the value is None.
            self.parameters = {
                k: v for k, v in merged.items() if v is not None
            }


# ** mapper: flagged_dependency_yaml_object
class FlaggedDependencyYamlObject(FlaggedDependency, TransferObject):
    '''
    A YAML data representation of a flagged dependency object.
    '''

    class Options:
        '''
        The options for the flagged dependency data.
        '''

        serialize_when_none = False
        roles = {
            'to_model': TransferObject.deny(),
            'to_data.yaml': TransferObject.deny('flag'),
        }

    # * attribute: parameters
    parameters = DictType(
        StringType,
        default={},
        serialized_name='params',
        deserialize_from=['params', 'parameters'],
        metadata=dict(
            description='The parameters for the dependency.'
        )
    )

    # * method: map
    def map(self, flag: str = None, **kwargs) -> FlaggedDependency:
        '''
        Maps the flagged dependency data to a flagged dependency object.

        :param flag: The flag for the dependency.
        :type flag: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new flagged dependency object.
        :rtype: FlaggedDependency
        '''

        # Map to the flagged dependency object.
        # Note: parent map() already calls to_primitive, so we only pass overrides.
        return super().map(
            FlaggedDependency,
            flag=flag or self.flag,
            parameters=self.parameters,
            **kwargs
        )

    # * method: from_model
    @staticmethod
    def from_model(flagged_dependency: FlaggedDependency, **kwargs) -> 'FlaggedDependencyYamlObject':
        '''
        Creates a FlaggedDependencyYamlObject from a FlaggedDependency model.

        :param flagged_dependency: The flagged dependency model.
        :type flagged_dependency: FlaggedDependency
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new FlaggedDependencyYamlObject.
        :rtype: FlaggedDependencyYamlObject
        '''

        # Create a new FlaggedDependencyYamlObject from the model.
        return TransferObject.from_model(
            FlaggedDependencyYamlObject,
            flagged_dependency,
            **kwargs,
        )


# ** mapper: service_configuration_aggregate
class ServiceConfigurationAggregate(ServiceConfiguration, Aggregate):
    '''
    An aggregate representation of a service configuration.
    '''

    # * method: new
    @staticmethod
    def new(
        service_configuration_data: Dict[str, Any],
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> 'ServiceConfigurationAggregate':
        '''
        Initializes a new service configuration aggregate.

        :param service_configuration_data: The data to create the service configuration aggregate from.
        :type service_configuration_data: dict
        :param validate: True to validate the aggregate object.
        :type validate: bool
        :param strict: True to enforce strict mode for the aggregate object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new service configuration aggregate.
        :rtype: ServiceConfigurationAggregate
        '''

        # Create a new service configuration aggregate from the provided data.
        return Aggregate.new(
            ServiceConfigurationAggregate,
            validate=validate,
            strict=strict,
            **service_configuration_data,
            **kwargs
        )

    # * method: set_default_type
    def set_default_type(
        self,
        module_path: str | None = None,
        class_name: str | None = None,
        parameters: Dict[str, Any] | None = None,
    ) -> None:
        '''
        Update the default type and parameters for this service configuration.

        :param module_path: New module path (or None to clear).
        :type module_path: str | None
        :param class_name: New class name (or None to clear).
        :type class_name: str | None
        :param parameters: New parameters dict (or None to clear all).
        :type parameters: Dict[str, Any] | None
        :return: None
        :rtype: None
        '''

        # If both type fields are None, clear default type entirely.
        if module_path is None and class_name is None:
            self.module_path = None
            self.class_name = None
            self.parameters = {}

        # Otherwise, set them to whatever is provided.
        else:
            self.module_path = module_path
            self.class_name = class_name

        # Update parameters: if parameters is None, clear all; otherwise replace with filtered dict.
        if parameters is None:
            self.parameters = {}
        else:
            self.parameters = {k: v for k, v in parameters.items() if v is not None}

    # * method: set_dependency
    def set_dependency(
        self,
        flag: str,
        module_path: str | None = None,
        class_name: str | None = None,
        parameters: Dict[str, Any] | None = None,
    ) -> None:
        '''
        Sets or updates a flagged dependency.

        :param flag: The flag that identifies the dependency.
        :type flag: str
        :param module_path: The module path for the dependency.
        :type module_path: str | None
        :param class_name: The class name for the dependency.
        :type class_name: str | None
        :param parameters: The parameters for the dependency (empty dict by default).
        :type parameters: Dict[str, Any] | None
        :return: None
        :rtype: None
        '''

        # Normalize parameters to a dict.
        parameters = parameters or {}

        # Replace the value of the dependency if a dependency with the same flag exists.
        for dep in self.dependencies:
            if dep.flag == flag:
                dep.module_path = module_path
                dep.class_name = class_name

                # Inline set_parameters semantics — works on any FlaggedDependency instance.
                merged = dict(dep.parameters or {})
                merged.update(parameters)
                dep.parameters = {k: v for k, v in merged.items() if v is not None}
                return

        # Create a new dependency if none exists with this flag.
        dependency = DomainObject.new(
            FlaggedDependency,
            module_path=module_path,
            class_name=class_name,
            flag=flag,
            parameters=parameters,
        )

        self.dependencies.append(dependency)

    # * method: remove_dependency
    def remove_dependency(self, flag: str) -> None:
        '''
        Remove a flagged dependency by its flag.

        :param flag: The flag identifying the dependency to remove.
        :type flag: str
        :return: None
        :rtype: None
        '''

        # Filter out any dependency whose flag matches the provided flag.
        self.dependencies = [
            dependency
            for dependency in self.dependencies
            if dependency.flag != flag
        ]


# ** mapper: service_configuration_yaml_object
class ServiceConfigurationYamlObject(ServiceConfiguration, TransferObject):
    '''
    A YAML data representation of a service configuration object.
    '''

    class Options:
        '''
        The options for the service configuration data.
        '''

        serialize_when_none = False
        roles = {
            'to_model': TransferObject.deny('dependencies', 'parameters'),
            'to_data.yaml': TransferObject.deny('id'),
        }

    # * attribute: dependencies
    dependencies = DictType(
        ModelType(FlaggedDependencyYamlObject),
        default={},
        serialized_name='deps',
        deserialize_from=['deps', 'dependencies', 'flags'],
        metadata=dict(
            description='The dependencies as key-value pairs, keyed by flags.'
        )
    )

    # * attribute: parameters
    parameters = DictType(
        StringType,
        default={},
        serialized_name='params',
        deserialize_from=['params', 'parameters'],
        metadata=dict(
            description='The default parameters for the service configuration.'
        )
    )

    # * method: map
    def map(self, **kwargs) -> ServiceConfigurationAggregate:
        '''
        Maps the service configuration data to a service configuration aggregate.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new service configuration aggregate.
        :rtype: ServiceConfigurationAggregate
        '''

        # Map the service configuration data to a service configuration aggregate.
        return super().map(
            ServiceConfigurationAggregate,
            dependencies=[dep.map(flag=flag) for flag, dep in self.dependencies.items()],
            parameters=self.parameters,
            **self.to_primitive('to_model'),
            **kwargs
        )

    # * method: from_model
    @staticmethod
    def from_model(service_configuration: ServiceConfiguration, **kwargs) -> 'ServiceConfigurationYamlObject':
        '''
        Creates a ServiceConfigurationYamlObject from a ServiceConfiguration model.

        :param service_configuration: The service configuration model.
        :type service_configuration: ServiceConfiguration
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new ServiceConfigurationYamlObject.
        :rtype: ServiceConfigurationYamlObject
        '''

        # Create a new ServiceConfigurationYamlObject from the model, converting
        # the dependencies list into a dictionary keyed by flag.
        return TransferObject.from_model(
            ServiceConfigurationYamlObject,
            service_configuration,
            dependencies={
                dep.flag: TransferObject.from_model(FlaggedDependencyYamlObject, dep)
                for dep in service_configuration.dependencies
            },
            **kwargs,
        )



"""Tiferet DI Domain Events"""

# *** imports

# ** core
from typing import Tuple, List, Dict, Any, Optional

# ** app
from .settings import DomainEvent, a
from ..domain.di import ServiceConfiguration
from ..interfaces.di import DIService
from ..mappers.di import ServiceConfigurationAggregate

# *** events

# ** event: add_service_configuration
class AddServiceConfiguration(DomainEvent):
    '''
    A domain event to add a new service configuration.
    '''

    # * attribute: di_service
    di_service: DIService

    # * init
    def __init__(self, di_service: DIService):
        '''
        Initialize the AddServiceConfiguration event.

        :param di_service: The DI service.
        :type di_service: DIService
        '''

        # Set the event attributes.
        self.di_service = di_service

    # * method: execute
    @DomainEvent.parameters_required(['id'])
    def execute(
        self,
        id: str,
        module_path: Optional[str] = None,
        class_name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = {},
        flagged_dependencies: Optional[List[Dict[str, Any]]] = [],
        **kwargs,
    ) -> ServiceConfiguration:
        '''
        Add a new service configuration.

        :param id: Required unique identifier.
        :type id: str
        :param module_path: Optional default module path.
        :type module_path: str | None
        :param class_name: Optional default class name.
        :type class_name: str | None
        :param parameters: Optional configuration parameters (default {}).
        :type parameters: Dict[str, Any] | None
        :param flagged_dependencies: Optional list of flagged dependencies (default []).
        :type flagged_dependencies: List[Dict[str, Any]] | None
        :return: Created ServiceConfiguration model.
        :rtype: ServiceConfiguration
        '''

        # Check for existing configuration id.
        self.verify(
            not self.di_service.configuration_exists(id),
            a.const.CONFIGURATION_ALREADY_EXISTS_ID,
            id=id,
        )

        # Validate at least one type source (default type or flagged dependencies).
        has_default = bool(module_path and class_name)
        has_deps = bool(flagged_dependencies)
        self.verify(
            has_default or has_deps,
            a.const.INVALID_SERVICE_CONFIGURATION_ID,
        )

        # Create service configuration aggregate from dependency dicts.
        configuration = ServiceConfigurationAggregate.new(
            service_configuration_data=dict(
                id=id,
                module_path=module_path,
                class_name=class_name,
                parameters=parameters,
                dependencies=flagged_dependencies,
            ),
        )

        # Save the new configuration and return it.
        self.di_service.save_configuration(configuration)
        return configuration

# ** event: set_default_service_configuration
class SetDefaultServiceConfiguration(DomainEvent):
    '''
    A domain event to set or update the default service configuration for an
    existing service configuration.
    '''

    # * attribute: di_service
    di_service: DIService

    # * init
    def __init__(self, di_service: DIService):
        '''
        Initialize the SetDefaultServiceConfiguration event.

        :param di_service: The DI service.
        :type di_service: DIService
        '''

        # Set the event attributes.
        self.di_service = di_service

    # * method: execute
    def execute(
        self,
        id: str,
        module_path: Optional[str] = None,
        class_name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> ServiceConfiguration:
        '''
        Set or update the default module/class and parameters for an
        existing service configuration.

        :param id: The unique service configuration identifier.
        :type id: str
        :param module_path: Optional default module path.
        :type module_path: str | None
        :param class_name: Optional default class name.
        :type class_name: str | None
        :param parameters: Optional configuration parameters. When None,
            existing parameters are cleared.
        :type parameters: Dict[str, Any] | None
        :return: The updated service configuration.
        :rtype: ServiceConfiguration
        '''

        # Retrieve the existing configuration.
        configuration = self.di_service.get_configuration(id)

        # Verify that the configuration exists.
        self.verify(
            configuration is not None,
            a.const.SERVICE_CONFIGURATION_NOT_FOUND_ID,
            id=id,
        )

        # If either module_path or class_name is provided, both must be
        # non-None to ensure an atomic default type update.
        if module_path is not None or class_name is not None:
            self.verify(
                module_path is not None and class_name is not None,
                a.const.INVALID_SERVICE_CONFIGURATION_ID,
            )

            # Update both type and parameters via the model helper.
            configuration.set_default_type(
                module_path,
                class_name,
                parameters,
            )
        else:
            # Only parameters are being updated (or cleared). Keep the
            # existing module_path and class_name but delegate parameter
            # handling/cleanup to the model helper.
            configuration.set_default_type(
                configuration.module_path,
                configuration.class_name,
                parameters,
            )

        # Persist the updated configuration and return it.
        self.di_service.save_configuration(configuration)
        return configuration

# ** event: set_service_dependency
class SetServiceDependency(DomainEvent):
    '''
    A domain event to set or update a flagged dependency on an existing
    service configuration.
    '''

    # * attribute: di_service
    di_service: DIService

    # * init
    def __init__(self, di_service: DIService):
        '''
        Initialize the SetServiceDependency event.

        :param di_service: The DI service.
        :type di_service: DIService
        '''

        # Set the event attributes.
        self.di_service = di_service

    # * method: execute
    @DomainEvent.parameters_required(['flag'])
    def execute(
        self,
        id: str,
        flag: str,
        module_path: str,
        class_name: str,
        parameters: Dict[str, Any] = {},
        **kwargs,
    ) -> str:
        '''
        Set or update a flagged dependency for the given service
        configuration.

        :param id: The service configuration identifier.
        :type id: str
        :param flag: The flag that identifies the dependency.
        :type flag: str
        :param module_path: The module path for the dependency.
        :type module_path: str
        :param class_name: The class name for the dependency.
        :type class_name: str
        :param parameters: Parameters for the dependency.
        :type parameters: Dict[str, Any]
        :return: The service configuration id.
        :rtype: str
        '''

        # Ensure module_path and class_name are both provided for a valid
        # flagged dependency.
        self.verify(
            bool(module_path) and bool(class_name),
            a.const.INVALID_FLAGGED_DEPENDENCY_ID,
        )

        # Retrieve the existing configuration.
        configuration = self.di_service.get_configuration(id)

        # Verify that the configuration exists.
        self.verify(
            configuration is not None,
            a.const.SERVICE_CONFIGURATION_NOT_FOUND_ID,
            id=id,
        )

        # Delegate to the model to set/update the flagged dependency.
        configuration.set_dependency(
            flag=flag,
            module_path=module_path,
            class_name=class_name,
            parameters=parameters,
        )

        # Persist the updated configuration.
        self.di_service.save_configuration(configuration)

        # Return the id for convenience/confirmation.
        return id


# ** event: remove_service_dependency
class RemoveServiceDependency(DomainEvent):
    '''
    A domain event to remove a flagged dependency from an existing service
    configuration.
    '''

    # * attribute: di_service
    di_service: DIService

    # * init
    def __init__(self, di_service: DIService):
        '''
        Initialize the RemoveServiceDependency event.

        :param di_service: The DI service.
        :type di_service: DIService
        '''

        # Set the event attributes.
        self.di_service = di_service

    # * method: execute
    @DomainEvent.parameters_required(['flag'])
    def execute(
        self,
        id: str,
        flag: str,
        **kwargs,
    ) -> str:
        '''
        Remove a flagged dependency from the given service configuration.

        :param id: The service configuration identifier.
        :type id: str
        :param flag: The flag that identifies the dependency to remove.
        :type flag: str
        :return: The service configuration id.
        :rtype: str
        '''

        # Retrieve the existing configuration.
        configuration = self.di_service.get_configuration(id)

        # Verify that the configuration exists.
        self.verify(
            configuration is not None,
            a.const.SERVICE_CONFIGURATION_NOT_FOUND_ID,
            id=id,
        )

        # Remove the dependency by flag (idempotent at the model level).
        configuration.remove_dependency(flag)

        # Post-removal validation: ensure a remaining type source.
        has_default = bool(configuration.module_path and configuration.class_name)
        has_deps = bool(configuration.dependencies)
        self.verify(
            has_default or has_deps,
            a.const.INVALID_SERVICE_CONFIGURATION_ID,
        )

        # Persist the updated configuration.
        self.di_service.save_configuration(configuration)

        # Return the id for convenience/confirmation.
        return id

# ** event: remove_service_configuration
class RemoveServiceConfiguration(DomainEvent):
    '''
    A domain event to remove a service configuration by ID.
    '''

    # * attribute: di_service
    di_service: DIService

    # * init
    def __init__(self, di_service: DIService):
        '''
        Initialize the RemoveServiceConfiguration event.

        :param di_service: The DI service.
        :type di_service: DIService
        '''

        # Set the event attributes.
        self.di_service = di_service

    # * method: execute
    @DomainEvent.parameters_required(['id'])
    def execute(self, id: str, **kwargs) -> str:
        '''
        Remove a service configuration.

        :param id: The unique identifier of the configuration to remove.
        :type id: str
        :param kwargs: Additional context.
        :type kwargs: dict
        :return: The removed configuration ID.
        :rtype: str
        '''

        # Delete (idempotent; underlying service handles non-existent IDs).
        self.di_service.delete_configuration(id)

        # Return id for confirmation.
        return id


# ** event: set_service_constants
class SetServiceConstants(DomainEvent):
    '''
    A domain event to set or clear service-level constants.
    '''

    # * attribute: di_service
    di_service: DIService

    # * init
    def __init__(self, di_service: DIService):
        '''
        Initialize the SetServiceConstants event.

        :param di_service: The DI service to use.
        :type di_service: DIService
        '''

        # Set the event attributes.
        self.di_service = di_service

    # * method: execute
    def execute(
        self,
        constants: Optional[Dict[str, Any]] = {},
        **kwargs,
    ) -> Dict[str, Any]:
        '''
        Set service constants.

        :param constants: New constants dictionary, or None to clear all.
            Keys with None value are removed using pop-style semantics.
        :type constants: Dict[str, Any] | None
        :param kwargs: Additional context.
        :type kwargs: dict
        :return: The updated constants dictionary.
        :rtype: Dict[str, Any]
        '''

        # Retrieve current constants from the DI service.
        _, current_constants = self.di_service.list_all()

        if constants is None:
            # Clear all constants.
            updated = {}
        else:
            # Start from existing constants, then apply updates/removals.
            updated = dict(current_constants or {})

            # Update the current constants giving preference to the added ones.
            updated.update(constants)

            # Remove any constants with a value of None.
            updated = {k: v for k, v in updated.items() if v != None}

        # Persist the updated constants.
        self.di_service.save_constants(updated)

        # Return the updated constants dictionary.
        return updated


# ** event: list_all_settings
class ListAllSettings(DomainEvent):
    '''
    A domain event to list all service configurations and constants.
    '''

    # * attribute: di_service
    di_service: DIService

    # * init
    def __init__(self, di_service: DIService):
        '''
        Initialize the ListAllSettings event.

        :param di_service: The DI service.
        :type di_service: DIService
        '''

        # Set the event attributes.
        self.di_service = di_service

    # * method: execute
    def execute(self) -> Tuple[List[ServiceConfiguration], Dict[str, Any]]:
        '''
        Execute the list all settings event.

        :return: The list of all service configurations and constants.
        :rtype: Tuple[List[ServiceConfiguration], Dict[str, Any]]
        '''

        # Return the list of all service configurations from the DI service.
        return self.di_service.list_all()



"""Tiferet Feature Mappers"""

# *** imports

# ** core
from typing import Dict, Any, List

# ** app
from ..domain import (
    Feature,
    FeatureStep,
    FeatureEvent,
    ListType,
    ModelType,
    DictType,
    StringType,
)
from ..events import RaiseError, a
from .settings import (
    Aggregate,
    TransferObject,
)

# *** mappers

# ** mapper: feature_event_aggregate
class FeatureEventAggregate(FeatureEvent, Aggregate):
    '''
    An aggregate representation of a feature event.
    '''

    # * method: new
    @staticmethod
    def new(
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> 'FeatureEventAggregate':
        '''
        Initializes a new feature event aggregate.

        :param validate: True to validate the aggregate object.
        :type validate: bool
        :param strict: True to enforce strict mode for the aggregate object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new feature event aggregate.
        :rtype: FeatureEventAggregate
        '''

        # Create a new feature event aggregate from the provided data.
        return Aggregate.new(
            FeatureEventAggregate,
            validate=validate,
            strict=strict,
            **kwargs
        )

    # * method: set_pass_on_error
    def set_pass_on_error(self, value: Any) -> None:
        '''
        Set the ``pass_on_error`` flag based on a provided value.

        :param value: The value to interpret as a boolean.
        :type value: Any
        :return: None
        :rtype: None
        '''

        # Normalize the value, treating the string "false" (case-insensitive)
        # as an explicit False value and using standard bool conversion otherwise.
        if isinstance(value, str) and value.lower() == 'false':
            self.pass_on_error = False
        else:
            self.pass_on_error = bool(value)

    # * method: set_parameters
    def set_parameters(self, parameters: Dict[str, Any] | None = None) -> None:
        '''
        Merge new parameters into the existing parameters, preferring new
        values and removing keys with ``None`` values.

        :param parameters: The new parameters to merge.
        :type parameters: dict | None
        :return: None
        :rtype: None
        '''

        # Do nothing if no parameters were provided.
        if parameters is None:
            return

        # Start from the existing parameters and update with new values.
        merged = dict(self.parameters or {})
        merged.update(parameters)

        # Remove any keys whose value is None.
        self.parameters = {k: v for k, v in merged.items() if v is not None}

    # * method: set_attribute
    def set_attribute(self, attribute: str, value: Any) -> None:
        '''
        Set an attribute on the feature command, with special handling for
        ``parameters`` and ``pass_on_error``.

        :param attribute: The attribute name to set.
        :type attribute: str
        :param value: The value to apply to the attribute.
        :type value: Any
        :return: None
        :rtype: None
        '''

        # Delegate to specialized helpers for parameters and pass_on_error.
        if attribute == 'parameters':
            self.set_parameters(value)
        elif attribute == 'pass_on_error':
            self.set_pass_on_error(value)
        else:
            setattr(self, attribute, value)


# ** mapper: feature_event_yaml_object
class FeatureEventYamlObject(FeatureEvent, TransferObject):
    '''
    A YAML data representation of a feature event object.
    '''

    class Options():
        '''
        The options for the feature event data.
        '''

        serialize_when_none = False
        roles = {
            'to_model': TransferObject.deny('type'),
            'to_data.yaml': TransferObject.deny('type'),
        }

    # * attribute: parameters
    parameters = DictType(
        StringType(),
        default={},
        serialized_name='params',
        deserialize_from=['params', 'parameters'],
        metadata=dict(
            description='The parameters for the feature event.'
        )
    )

    # * method: map
    def map(self, **kwargs) -> FeatureEvent:
        '''
        Maps the feature event data to a feature event object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new feature event object.
        :rtype: FeatureEvent
        '''

        # Map to the feature event aggregate.
        return super().map(
            FeatureEventAggregate,
            parameters=self.parameters,
            **self.to_primitive('to_model'),
            **kwargs
        )

    # * method: from_model
    @staticmethod
    def from_model(feature_event: FeatureEvent, **kwargs) -> 'FeatureEventYamlObject':
        '''
        Creates a FeatureEventYamlObject from a FeatureEvent model.

        :param feature_event: The feature event model.
        :type feature_event: FeatureEvent
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new FeatureEventYamlObject.
        :rtype: FeatureEventYamlObject
        '''

        # Create a new FeatureEventYamlObject from the model.
        return TransferObject.from_model(
            FeatureEventYamlObject,
            feature_event,
            **kwargs,
        )

# ** mapper: feature_aggregate
class FeatureAggregate(Feature, Aggregate):
    '''
    An aggregate representation of a feature.
    '''

    # * method: new
    @staticmethod
    def new(
        name: str = None,
        group_id: str = None,
        feature_key: str = None,
        id: str = None,
        description: str = None,
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> 'FeatureAggregate':
        '''
        Initializes a new feature aggregate.

        :param name: The name of the feature.
        :type name: str
        :param group_id: The context group identifier of the feature.
        :type group_id: str
        :param feature_key: The key of the feature.
        :type feature_key: str
        :param id: The identifier of the feature.
        :type id: str
        :param description: The description of the feature.
        :type description: str
        :param validate: True to validate the aggregate object.
        :type validate: bool
        :param strict: True to enforce strict mode for the aggregate object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new feature aggregate.
        :rtype: FeatureAggregate
        '''

        # Derive group_id and feature_key from id if provided.
        if id and '.' in id and (not group_id or not feature_key):
            group_id, feature_key = id.split('.', 1)

        # Set the feature key as the snake case of the name if not provided.
        if name and not feature_key:
            feature_key = name.lower().replace(' ', '_')

        # Feature ID is the group ID and feature key separated by a period.
        if not id and group_id and feature_key:
            id = f'{group_id}.{feature_key}'

        # Set the description as the name if not provided.
        if name and not description:
            description = name

        # Create a new feature aggregate from the provided data.
        return Aggregate.new(
            FeatureAggregate,
            validate=validate,
            strict=strict,
            id=id,
            name=name,
            group_id=group_id,
            feature_key=feature_key,
            description=description,
            **kwargs
        )

    # * method: add_step
    def add_step(
        self,
        name: str,
        service_id: str,
        parameters: Dict[str, Any] | None = None,
        data_key: str | None = None,
        pass_on_error: bool = False,
        position: int | None = None,
    ) -> FeatureEvent:
        '''
        Add a feature event step using raw attributes.

        :param name: Step name.
        :type name: str
        :param service_id: Service configuration ID.
        :type service_id: str
        :param parameters: Optional parameters dictionary.
        :type parameters: dict | None
        :param data_key: Optional result data key.
        :type data_key: str | None
        :param pass_on_error: Whether to pass on errors from this step.
        :type pass_on_error: bool
        :param position: Insertion position (None to append).
        :type position: int | None
        :return: Created FeatureEvent instance.
        :rtype: FeatureEvent
        '''

        # Create the feature event from raw attributes.
        step = FeatureEventAggregate.new(
            name=name,
            service_id=service_id,
            parameters=parameters or {},
            data_key=data_key,
            pass_on_error=pass_on_error,
        )

        # Add the feature event step to the feature.
        if position is not None:
            self.steps.insert(position, step)
        else:
            self.steps.append(step)

        return step

    # * method: get_step
    def get_step(self, position: int) -> FeatureStep | None:
        '''
        Get the feature step at the given position, or ``None`` if the
        index is out of range or invalid.

        :param position: The index of the step to retrieve.
        :type position: int
        :return: The FeatureStep at the position, or None.
        :rtype: FeatureStep | None
        '''

        # Attempt to retrieve the step at the specified index.
        try:
            return self.steps[position]
        except (IndexError, TypeError):
            return None

    # * method: remove_step
    def remove_step(self, position: int) -> FeatureStep | None:
        '''
        Remove and return the feature step at the given position, or
        return ``None`` if the index is out of range or invalid.

        :param position: The index of the feature step to remove.
        :type position: int
        :return: The removed feature step or ``None``.
        :rtype: FeatureStep | None
        '''

        # Validate the position argument.
        if not isinstance(position, int) or position < 0:
            return None

        # Attempt to remove and return the step at the specified index.
        try:
            return self.steps.pop(position)
        except IndexError:
            return None

    # * method: reorder_step
    def reorder_step(self, current_position: int, new_position: int) -> FeatureStep | None:
        '''
        Move a feature step from its current position to a new position
        within the ``steps`` list.

        :param current_position: Current index of the step.
        :type current_position: int
        :param new_position: Desired new index.
        :type new_position: int
        :return: Moved step or ``None`` if ``current_position`` is invalid.
        :rtype: FeatureStep | None
        '''

        # Attempt to remove the step at the current position.
        try:
            step = self.steps.pop(current_position)
        except (IndexError, TypeError):
            return None

        # Clamp the new position index to the valid range.
        if new_position < 0:
            new_position = 0
        if new_position > len(self.steps):
            new_position = len(self.steps)

        # Insert the step at the clamped position and return it.
        self.steps.insert(new_position, step)

        return step

    # * method: rename
    def rename(self, name: str) -> None:
        '''
        Update the display name of the feature.

        :param name: The new name.
        :type name: str
        :return: None
        :rtype: None
        '''

        self.name = name

        # Perform final aggregate validation.
        self.validate()

    # * method: set_description
    def set_description(self, description: str | None) -> None:
        '''
        Update the feature description.

        :param description: The new description.
        :type description: str | None
        :return: None
        :rtype: None
        '''

        self.description = description


# ** mapper: feature_yaml_object
class FeatureYamlObject(Feature, TransferObject):
    '''
    A YAML data representation of a feature object.
    '''

    class Options():
        '''
        The options for the feature data.
        '''

        serialize_when_none = False
        roles = {
            'to_model': TransferObject.deny('steps'),
            'to_data.yaml': TransferObject.deny('feature_key', 'group_id', 'id'),
        }

    # * attribute: steps
    steps = ListType(
        ModelType(FeatureEventYamlObject),
        deserialize_from=['handlers', 'functions', 'commands', 'steps'],
        default=[],
    )

    # * method: map
    def map(self, **kwargs) -> FeatureAggregate:
        '''
        Maps the feature data to a feature aggregate.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new feature aggregate.
        :rtype: FeatureAggregate
        '''

        # Map the feature data.
        return super().map(
            FeatureAggregate,
            steps=[step.map() for step in (self.steps or [])],
            **self.to_primitive('to_model'),
            **kwargs
        )

    # * method: from_model
    @staticmethod
    def from_model(feature: Feature, **kwargs) -> 'FeatureYamlObject':
        '''
        Creates a FeatureYamlObject from a Feature model.

        :param feature: The feature model.
        :type feature: Feature
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new FeatureYamlObject.
        :rtype: FeatureYamlObject
        '''

        # Create a new FeatureYamlObject from the model, converting
        # the steps list into FeatureEventYamlObject instances.
        return TransferObject.from_model(
            FeatureYamlObject,
            feature,
            steps=[
                TransferObject.from_model(FeatureEventYamlObject, step)
                for step in feature.steps
            ],
            **kwargs,
        )



"""Tiferet Logging Events"""

# *** imports

# ** core
from typing import List, Tuple

# ** app
from ..domain import Formatter, Handler, Logger
from ..interfaces import LoggingService
from ..mappers import FormatterAggregate, HandlerAggregate, LoggerAggregate
from .settings import DomainEvent, a


# *** events

# ** event: list_all_logging_configs
class ListAllLoggingConfigs(DomainEvent):
    '''
    Event to list all logging configurations (formatters, handlers, loggers).
    '''

    # * attribute: logging_service
    logging_service: LoggingService

    # * init
    def __init__(self, logging_service: LoggingService):
        '''
        Initialize the ListAllLoggingConfigs event.

        :param logging_service: The logging service to use.
        :type logging_service: LoggingService
        '''

        # Set the logging service dependency.
        self.logging_service = logging_service

    # * method: execute
    def execute(self, **kwargs) -> Tuple[List[Formatter], List[Handler], List[Logger]]:
        '''
        List all logging configurations.

        :param kwargs: Additional keyword arguments (unused).
        :type kwargs: dict
        :return: A tuple of (formatters, handlers, loggers).
        :rtype: Tuple[List[Formatter], List[Handler], List[Logger]]
        '''

        # Delegate to the logging service.
        return self.logging_service.list_all()


# ** event: add_formatter
class AddFormatter(DomainEvent):
    '''
    Event to add a new logging formatter configuration.
    '''

    # * attribute: logging_service
    logging_service: LoggingService

    # * init
    def __init__(self, logging_service: LoggingService):
        '''
        Initialize the AddFormatter event.

        :param logging_service: The logging service to use.
        :type logging_service: LoggingService
        '''

        # Set the logging service dependency.
        self.logging_service = logging_service

    # * method: execute
    @DomainEvent.parameters_required(['id', 'name', 'format'])
    def execute(
            self,
            id: str,
            name: str,
            format: str,
            description: str | None = None,
            datefmt: str | None = None,
            **kwargs,
        ) -> Formatter:
        '''
        Add a new formatter.

        :param id: The unique identifier for the formatter.
        :type id: str
        :param name: The name of the formatter.
        :type name: str
        :param format: The format string for log messages.
        :type format: str
        :param description: Optional description of the formatter.
        :type description: str | None
        :param datefmt: Optional date format string.
        :type datefmt: str | None
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The created Formatter aggregate.
        :rtype: Formatter
        '''

        # Create the formatter aggregate.
        formatter = FormatterAggregate.new(
            id=id,
            name=name,
            format=format,
            description=description,
            datefmt=datefmt,
            **kwargs,
        )

        # Persist the formatter.
        self.logging_service.save_formatter(formatter)

        # Return the created formatter.
        return formatter


# ** event: remove_formatter
class RemoveFormatter(DomainEvent):
    '''
    Event to remove a formatter configuration by ID (idempotent).
    '''

    # * attribute: logging_service
    logging_service: LoggingService

    # * init
    def __init__(self, logging_service: LoggingService):
        '''
        Initialize the RemoveFormatter event.

        :param logging_service: The logging service to use.
        :type logging_service: LoggingService
        '''

        # Set the logging service dependency.
        self.logging_service = logging_service

    # * method: execute
    @DomainEvent.parameters_required(['id'])
    def execute(self, id: str, **kwargs) -> str:
        '''
        Remove a formatter by ID.

        :param id: The formatter ID.
        :type id: str
        :param kwargs: Additional keyword arguments (unused).
        :type kwargs: dict
        :return: The removed formatter ID.
        :rtype: str
        '''

        # Delete the formatter (idempotent operation).
        self.logging_service.delete_formatter(id)

        # Return the formatter ID.
        return id


# ** event: add_handler
class AddHandler(DomainEvent):
    '''
    Event to add a new logging handler configuration.
    '''

    # * attribute: logging_service
    logging_service: LoggingService

    # * init
    def __init__(self, logging_service: LoggingService):
        '''
        Initialize the AddHandler event.

        :param logging_service: The logging service to use.
        :type logging_service: LoggingService
        '''

        # Set the logging service dependency.
        self.logging_service = logging_service

    # * method: execute
    @DomainEvent.parameters_required(['id', 'name', 'module_path', 'class_name', 'level', 'formatter'])
    def execute(
            self,
            id: str,
            name: str,
            module_path: str,
            class_name: str,
            level: str,
            formatter: str,
            description: str | None = None,
            stream: str | None = None,
            filename: str | None = None,
            **kwargs,
        ) -> Handler:
        '''
        Add a new handler.

        :param id: The unique identifier for the handler.
        :type id: str
        :param name: The name of the handler.
        :type name: str
        :param module_path: The module path for the handler class.
        :type module_path: str
        :param class_name: The class name of the handler.
        :type class_name: str
        :param level: The logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL).
        :type level: str
        :param formatter: The formatter ID to use.
        :type formatter: str
        :param description: Optional description of the handler.
        :type description: str | None
        :param stream: Optional stream specification (e.g., ext://sys.stdout).
        :type stream: str | None
        :param filename: Optional filename for FileHandler.
        :type filename: str | None
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The created Handler aggregate.
        :rtype: Handler
        '''

        # Create the handler aggregate.
        handler = HandlerAggregate.new(
            id=id,
            name=name,
            module_path=module_path,
            class_name=class_name,
            level=level,
            formatter=formatter,
            description=description,
            stream=stream,
            filename=filename,
            **kwargs,
        )

        # Persist the handler.
        self.logging_service.save_handler(handler)

        # Return the created handler.
        return handler


# ** event: remove_handler
class RemoveHandler(DomainEvent):
    '''
    Event to remove a handler configuration by ID (idempotent).
    '''

    # * attribute: logging_service
    logging_service: LoggingService

    # * init
    def __init__(self, logging_service: LoggingService):
        '''
        Initialize the RemoveHandler event.

        :param logging_service: The logging service to use.
        :type logging_service: LoggingService
        '''

        # Set the logging service dependency.
        self.logging_service = logging_service

    # * method: execute
    @DomainEvent.parameters_required(['id'])
    def execute(self, id: str, **kwargs) -> str:
        '''
        Remove a handler by ID.

        :param id: The handler ID.
        :type id: str
        :param kwargs: Additional keyword arguments (unused).
        :type kwargs: dict
        :return: The removed handler ID.
        :rtype: str
        '''

        # Delete the handler (idempotent operation).
        self.logging_service.delete_handler(id)

        # Return the handler ID.
        return id


# ** event: add_logger
class AddLogger(DomainEvent):
    '''
    Event to add a new logger configuration.
    '''

    # * attribute: logging_service
    logging_service: LoggingService

    # * init
    def __init__(self, logging_service: LoggingService):
        '''
        Initialize the AddLogger event.

        :param logging_service: The logging service to use.
        :type logging_service: LoggingService
        '''

        # Set the logging service dependency.
        self.logging_service = logging_service

    # * method: execute
    @DomainEvent.parameters_required(['id', 'name', 'level', 'handlers'])
    def execute(
            self,
            id: str,
            name: str,
            level: str,
            handlers: List[str],
            description: str | None = None,
            propagate: bool = True,
            **kwargs,
        ) -> Logger:
        '''
        Add a new logger.

        :param id: The unique identifier for the logger.
        :type id: str
        :param name: The name of the logger.
        :type name: str
        :param level: The logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL).
        :type level: str
        :param handlers: List of handler IDs to attach to this logger.
        :type handlers: List[str]
        :param description: Optional description of the logger.
        :type description: str | None
        :param propagate: Whether to propagate messages to parent loggers.
        :type propagate: bool
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The created Logger aggregate.
        :rtype: Logger
        '''

        # Create the logger aggregate.
        logger = LoggerAggregate.new(
            id=id,
            name=name,
            level=level,
            handlers=handlers,
            description=description,
            propagate=propagate,
            **kwargs,
        )

        # Persist the logger.
        self.logging_service.save_logger(logger)

        # Return the created logger.
        return logger


# ** event: remove_logger
class RemoveLogger(DomainEvent):
    '''
    Event to remove a logger configuration by ID (idempotent).
    '''

    # * attribute: logging_service
    logging_service: LoggingService

    # * init
    def __init__(self, logging_service: LoggingService):
        '''
        Initialize the RemoveLogger event.

        :param logging_service: The logging service to use.
        :type logging_service: LoggingService
        '''

        # Set the logging service dependency.
        self.logging_service = logging_service

    # * method: execute
    @DomainEvent.parameters_required(['id'])
    def execute(self, id: str, **kwargs) -> str:
        '''
        Remove a logger by ID.

        :param id: The logger ID.
        :type id: str
        :param kwargs: Additional keyword arguments (unused).
        :type kwargs: dict
        :return: The removed logger ID.
        :rtype: str
        '''

        # Delete the logger (idempotent operation).
        self.logging_service.delete_logger(id)

        # Return the logger ID.
        return id



"""Tiferet Events Settings"""

# *** imports

# ** core
from typing import Dict, Any

# ** app
from ..assets import TiferetError
from .. import assets as a

# *** classes

# ** class: domain_event
class DomainEvent(object):
    '''
    A base class for a domain event object.
    '''

    # * method: execute
    def execute(self, **kwargs) -> Any:
        '''
        Execute the domain event.

        :param kwargs: The event arguments.
        :type kwargs: dict
        :return: The event result.
        :rtype: Any
        '''

        # Not implemented.
        raise NotImplementedError()

    # * method: raise_error (static)
    @staticmethod
    def raise_error(error_code: str, message: str = None, **kwargs):
        '''
        Raise a structured TiferetError.

        :param error_code: The error code.
        :type error_code: str
        :param message: The error message.
        :type message: str
        :param kwargs: Additional error keyword arguments.
        :type kwargs: dict
        '''

        # Raise the TiferetError with the given error code and arguments.
        raise TiferetError(
            error_code,
            message,
            **kwargs
        )

    # * method: verify
    def verify(self, expression: bool, error_code: str, message: str = None, **kwargs):
        '''
        Verify an expression and raise an error if it is false.

        :param expression: The expression to verify.
        :type expression: bool
        :param error_code: The error code.
        :type error_code: str
        :param message: The error message.
        :type message: str
        :param kwargs: Additional error keyword arguments.
        :type kwargs: dict
        '''

        # Verify the expression.
        try:
            assert expression
        except AssertionError:
            self.raise_error(
                error_code,
                message,
                **kwargs
            )

    # * method: parameters_required (static)
    @staticmethod
    def parameters_required(param_names: list):
        '''
        Declarative parameter validator decorator.

        Inspects kwargs for required parameters and raises a single
        aggregated TiferetError if any are missing, None, or empty strings.

        :param param_names: The list of required parameter names.
        :type param_names: list
        :return: The decorator function.
        :rtype: callable
        '''

        def decorator(method):
            def wrapper(self, *args, **kwargs):

                # Collect all missing or invalid parameters.
                missing = []
                for name in param_names:
                    if name not in kwargs:
                        missing.append(name)
                        continue

                    value = kwargs[name]

                    # None is invalid.
                    if value is None:
                        missing.append(name)
                        continue

                    # Empty or whitespace-only strings are invalid.
                    if isinstance(value, str) and not value.strip():
                        missing.append(name)
                        continue

                # Raise a single error with all violations if any found.
                if missing:
                    DomainEvent.raise_error(
                        a.const.COMMAND_PARAMETER_REQUIRED_ID,
                        message=f'Required parameters missing for {self.__class__.__name__}.',
                        parameters=missing,
                        command=self.__class__.__name__,
                    )

                # Call the wrapped method.
                return method(self, *args, **kwargs)

            return wrapper

        return decorator

    # * method: handle (static)
    @staticmethod
    def handle(
            event_cls: type,
            dependencies: Dict[str, Any] = {},
            **kwargs) -> Any:
        '''
        Handle a domain event instance via the instantiate-execute pattern.

        :param event_cls: The domain event class to handle.
        :type event_cls: type
        :param dependencies: The event dependencies.
        :type dependencies: Dict[str, Any]
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The result of the event.
        :rtype: Any
        '''

        # Instantiate the event with its dependencies.
        event_handler = event_cls(**dependencies)

        # Execute the event handler.
        result = event_handler.execute(**kwargs)
        return result



"""Tiferet Feature YAML Repository"""

# *** imports

# ** core
from typing import Any, Dict, List

# ** app
from ..interfaces import FeatureService
from ..mappers import (
    TransferObject,
    FeatureAggregate,
    FeatureYamlObject,
)
from ..utils import Yaml

# *** repos

# ** repo: feature_yaml_repository
class FeatureYamlRepository(FeatureService):
    '''
    The feature YAML repository.
    '''

    # * attribute: yaml_file
    yaml_file: str

    # * attribute: encoding
    encoding: str

    # * attribute: default_role
    default_role: str

    # * init
    def __init__(self, feature_yaml_file: str, encoding: str = 'utf-8') -> None:
        '''
        Initialize the feature YAML repository.

        :param feature_yaml_file: The YAML configuration file path.
        :type feature_yaml_file: str
        :param encoding: The file encoding (default is 'utf-8').
        :type encoding: str
        '''

        # Set the repository attributes.
        self.yaml_file = feature_yaml_file
        self.encoding = encoding
        self.default_role = 'to_data.yaml'

    # * method: exists
    def exists(self, id: str) -> bool:
        '''
        Check if a feature exists by ID.

        :param id: The feature identifier in the format "<group_id>.<feature_key>".
        :type id: str
        :return: True if the feature exists, otherwise False.
        :rtype: bool
        '''

        # Split the feature id into group and feature key.
        group_id, feature_key = id.split('.', 1)

        # Load the group-specific feature data from the configuration file.
        group_data: Dict[str, Any] = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load(
            start_node=lambda data: data.get('features', {}).get(group_id, None)
        )

        # If the group does not exist, return False.
        if not group_data:
            return False

        # Return whether the feature key exists in the group.
        return feature_key in group_data

    # * method: get
    def get(self, id: str) -> FeatureAggregate | None:
        '''
        Retrieve a feature by ID.

        :param id: The feature identifier in the format "<group_id>.<feature_key>".
        :type id: str
        :return: The feature aggregate or None if not found.
        :rtype: FeatureAggregate | None
        '''

        # Split the feature id into group and feature key.
        group_id, feature_key = id.split('.', 1)

        # Load the group-specific feature data from the configuration file.
        group_data: Dict[str, Any] = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load(
            start_node=lambda data: data.get('features', {}).get(group_id, None)
        )

        # If the group does not exist, return None.
        if not group_data:
            return None

        # Retrieve the specific feature data.
        feature_data = group_data.get(feature_key)

        # If the feature does not exist, return None.
        if not feature_data:
            return None

        # Map the feature data to a FeatureAggregate and return it.
        return TransferObject.from_data(
            FeatureYamlObject,
            id=f'{group_id}.{feature_key}',
            **feature_data,
        ).map()

    # * method: list
    def list(self, group_id: str | None = None) -> List[FeatureAggregate]:
        '''
        List all features, optionally filtered by group.

        :param group_id: Optional group identifier to filter features.
        :type group_id: str | None
        :return: A list of feature aggregates.
        :rtype: List[FeatureAggregate]
        '''

        # Load all groups and feature definitions from the configuration file.
        groups_data: Dict[str, Dict[str, Any]] = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load(
            start_node=lambda data: data.get('features', {})
        )

        # Initialize the list of FeatureYamlObject objects.
        features: List[FeatureYamlObject] = []

        # If a specific group is requested, limit to that group.
        if group_id:
            group_features = groups_data.get(group_id, {})
            for feature_key, feature_data in group_features.items():
                features.append(TransferObject.from_data(
                    FeatureYamlObject,
                    id=f'{group_id}.{feature_key}',
                    **feature_data,
                ))

        # Otherwise, flatten all groups.
        else:
            for group, group_features in groups_data.items():
                for feature_key, feature_data in group_features.items():
                    features.append(TransferObject.from_data(
                        FeatureYamlObject,
                        id=f'{group}.{feature_key}',
                        **feature_data,
                    ))

        # Map all FeatureYamlObject instances to FeatureAggregates and return them.
        return [feature.map() for feature in features]

    # * method: save
    def save(self, feature: FeatureAggregate) -> None:
        '''
        Save or update a feature.

        :param feature: The feature aggregate to save.
        :type feature: FeatureAggregate
        :return: None
        :rtype: None
        '''

        # Convert the feature model to a FeatureYamlObject.
        feature_data = FeatureYamlObject.from_model(feature)

        # Split the feature id for nested update.
        group_id, feature_key = feature.id.split('.', 1)

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load()

        # Update or insert the feature entry using nested setdefault.
        full_data.setdefault('features', {}).setdefault(group_id, {})[feature_key] = feature_data.to_primitive(self.default_role)

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding,
        ).save(data=full_data)

    # * method: delete
    def delete(self, id: str) -> None:
        '''
        Delete a feature by ID. This operation is idempotent.

        :param id: The feature identifier in the format "<group_id>.<feature_key>".
        :type id: str
        :return: None
        :rtype: None
        '''

        # Load all features data from the configuration file.
        features_data: Dict[str, Dict[str, Any]] = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load(
            start_node=lambda data: data.get('features', {})
        )

        # Split the feature id into group and feature key.
        group_id, feature_key = id.split('.', 1)

        # Retrieve the group data.
        group_data = features_data.get(group_id, {})

        # Pop the feature entry if it exists (idempotent).
        group_data.pop(feature_key, None)

        # If the group becomes empty, remove it from the features mapping.
        if not group_data and group_id in features_data:
            features_data.pop(group_id, None)
        else:
            features_data[group_id] = group_data

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load()

        # Update the features section.
        full_data['features'] = features_data

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding,
        ).save(data=full_data)



"""Tiferet CLI Mappers"""

# *** imports

# ** core
from typing import Dict, Any, List

# ** app
from ..domain import (
    CliCommand,
    CliArgument,
    DomainObject,
    StringType,
    ListType,
    ModelType,
)
from ..events import RaiseError, a
from .settings import (
    Aggregate,
    TransferObject,
)

# *** mappers

# ** mapper: cli_argument_aggregate
class CliArgumentAggregate(CliArgument, Aggregate):
    '''
    An aggregate representation of a CLI argument.
    '''

    # * method: new
    @staticmethod
    def new(
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> 'CliArgumentAggregate':
        '''
        Initializes a new CLI argument aggregate.

        :param validate: True to validate the aggregate object.
        :type validate: bool
        :param strict: True to enforce strict mode for the aggregate object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new CLI argument aggregate.
        :rtype: CliArgumentAggregate
        '''

        # Create a new CLI argument aggregate from the provided data.
        return Aggregate.new(
            CliArgumentAggregate,
            validate=validate,
            strict=strict,
            **kwargs
        )

    # * method: set_attribute
    def set_attribute(self, attribute: str, value: Any) -> None:
        '''
        Update a supported attribute on the CLI argument aggregate.

        Supported attributes: description, type, required, default, choices, nargs, action.

        :param attribute: The attribute name to update.
        :type attribute: str
        :param value: The new value.
        :type value: Any
        :return: None
        :rtype: None
        '''

        # Define the set of supported attributes.
        supported = {
            'description',
            'type',
            'required',
            'default',
            'choices',
            'nargs',
            'action',
        }

        # Validate the attribute name.
        if attribute not in supported:
            RaiseError.execute(
                error_code=a.const.INVALID_MODEL_ATTRIBUTE_ID,
                message='Invalid attribute: {attribute}. Supported attributes are {supported}.',
                attribute=attribute,
                supported=', '.join(sorted(supported)),
            )

        # Apply the update to the attribute.
        setattr(self, attribute, value)

        # Perform final aggregate validation.
        self.validate()


# ** mapper: cli_command_aggregate
class CliCommandAggregate(CliCommand, Aggregate):
    '''
    An aggregate representation of a CLI command.
    '''

    # * method: new
    @staticmethod
    def new(
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> 'CliCommandAggregate':
        '''
        Initializes a new CLI command aggregate.

        :param validate: True to validate the aggregate object.
        :type validate: bool
        :param strict: True to enforce strict mode for the aggregate object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new CLI command aggregate.
        :rtype: CliCommandAggregate
        '''

        # Create a new CLI command aggregate from the provided CLI command data.
        return Aggregate.new(
            CliCommandAggregate,
            validate=validate,
            strict=strict,
            **kwargs
        )

    # * method: add_argument
    def add_argument(self,
            name_or_flags: list,
            description: str = None,
            type: str = None,
            required: bool = None,
            default: str = None,
            choices: list = None,
            nargs: str = None,
            action: str = None
        ) -> None:
        '''
        Add an argument to the command.

        :param name_or_flags: The name or flags of the argument.
        :type name_or_flags: list
        :param description: A brief description of the argument.
        :type description: str
        :param type: The type of the argument (str, int, float).
        :type type: str
        :param required: Whether the argument is required.
        :type required: bool
        :param default: The default value of the argument.
        :type default: str
        :param choices: A list of valid choices for the argument.
        :type choices: list
        :param nargs: The number of arguments to consume.
        :type nargs: str
        :param action: The action to take when the argument is encountered.
        :type action: str
        :return: None
        :rtype: None
        '''

        # Create a new CliArgument instance using DomainObject.new.
        argument = DomainObject.new(
            CliArgument,
            name_or_flags=name_or_flags,
            description=description,
            type=type,
            required=required,
            default=default,
            choices=choices,
            nargs=nargs,
            action=action
        )

        # Append the argument to the command's arguments list.
        self.arguments.append(argument)

    # * method: set_attribute
    def set_attribute(self, attribute: str, value: Any) -> None:
        '''
        Update a supported scalar attribute on the CLI command aggregate.

        Supported attributes: name, description, key, group_key.

        :param attribute: The attribute name to update.
        :type attribute: str
        :param value: The new value.
        :type value: Any
        :return: None
        :rtype: None
        '''

        # Define the set of supported attributes.
        supported = {
            'name',
            'description',
            'key',
            'group_key',
        }

        # Validate the attribute name.
        if attribute not in supported:
            RaiseError.execute(
                error_code=a.const.INVALID_MODEL_ATTRIBUTE_ID,
                message='Invalid attribute: {attribute}. Supported attributes are {supported}.',
                attribute=attribute,
                supported=', '.join(sorted(supported)),
            )

        # Apply the update to the attribute.
        setattr(self, attribute, value)

        # Perform final aggregate validation.
        self.validate()


# ** mapper: cli_command_yaml_object
class CliCommandYamlObject(CliCommand, TransferObject):
    '''
    A YAML data representation of a CLI command object.
    '''

    class Options():
        '''
        The options for the CLI command data.
        '''

        serialize_when_none = False
        roles = {
            'to_model': TransferObject.deny('arguments'),
            'to_data.yaml': TransferObject.deny('id', 'arguments'),
        }

    # * attribute: arguments
    arguments = ListType(
        ModelType(CliArgument),
        default=[],
        deserialize_from=['args', 'arguments'],
        serialized_name='args',
        metadata={
            'description': 'The list of arguments for the CLI command.'
        },
    )

    # * method: to_primitive
    def to_primitive(self, role='to_data.yaml', **kwargs) -> Dict[str, Any]:
        '''
        Converts the CLI command data to a primitive dictionary representation.

        :param role: The role to use for the conversion.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A primitive dictionary representation of the CLI command data.
        :rtype: dict
        '''

        # Convert the CLI command data to a primitive dictionary, converting
        # the arguments list into dictionaries as well.
        return dict(
            **super().to_primitive(role=role, **kwargs),
            args=[
                arg.to_primitive()
                for arg in self.arguments
            ]
        )

    # * method: map
    def map(self, **kwargs) -> CliCommandAggregate:
        '''
        Maps the CLI command data to a CLI command aggregate.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new CLI command aggregate.
        :rtype: CliCommandAggregate
        '''

        # Map the CLI command data.
        return super().map(
            CliCommandAggregate,
            arguments=[
                arg.to_primitive()
                for arg in self.arguments
            ],
            **self.to_primitive('to_model'),
            **kwargs
        )

    # * method: from_model
    @staticmethod
    def from_model(cli_command: CliCommand, **kwargs) -> 'CliCommandYamlObject':
        '''
        Creates a CliCommandYamlObject from a CliCommand model.

        :param cli_command: The CLI command model.
        :type cli_command: CliCommand
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new CliCommandYamlObject.
        :rtype: CliCommandYamlObject
        '''

        # Create a new CliCommandYamlObject from the model, converting
        # the arguments list into primitive dictionaries.
        return TransferObject.from_model(
            CliCommandYamlObject,
            cli_command,
            arguments=[
                arg.to_primitive()
                for arg in cli_command.arguments
            ],
            **kwargs,
        )



"""Tiferet Domain Logging"""

# *** imports

# ** core
from typing import Any, Dict

# ** app
from .settings import (
    DomainObject,
    StringType,
    BooleanType,
    ListType,
)

# *** models

# ** model: formatter
class Formatter(DomainObject):
    '''
    An entity representing a logging formatter configuration.
    '''

    # * attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The unique identifier of the formatter.'
        )
    )

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the formatter.'
        )
    )

    # * attribute: description
    description = StringType(
        metadata=dict(
            description='The description of the formatter.'
        )
    )

    # * attribute: format
    format = StringType(
        required=True,
        metadata=dict(
            description='The format string for log messages.'
        )
    )

    # * attribute: datefmt
    datefmt = StringType(
        metadata=dict(
            description='The date format for log timestamps.'
        )
    )

    # * method: format_config
    def format_config(self) -> Dict[str, Any]:
        '''
        Format the formatter configuration into a dictionary.

        :return: The formatted formatter configuration.
        :rtype: Dict[str, Any]
        '''

        # Return the formatter configuration.
        return {
            'format': self.format,
            'datefmt': self.datefmt
        }


# ** model: handler
class Handler(DomainObject):
    '''
    An entity representing a logging handler configuration.
    '''

    # * attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The unique identifier of the handler.'
        )
    )

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the handler.'
        )
    )

    # * attribute: description
    description = StringType(
        metadata=dict(
            description='The description of the handler.'
        )
    )

    # * attribute: module_path
    module_path = StringType(
        required=True,
        metadata=dict(
            description='The module path for the handler class.'
        )
    )

    # * attribute: class_name
    class_name = StringType(
        required=True,
        metadata=dict(
            description='The class name of the handler.'
        )
    )

    # * attribute: level
    level = StringType(
        required=True,
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        metadata=dict(
            description='The logging level for the handler (e.g., INFO, DEBUG).'
        )
    )

    # * attribute: formatter
    formatter = StringType(
        required=True,
        metadata=dict(
            description='The id of the formatter to use.'
        )
    )

    # * attribute: stream
    stream = StringType(
        metadata=dict(
            description='The stream for StreamHandler (e.g., ext://sys.stdout).'
        )
    )

    # * attribute: filename
    filename = StringType(
        metadata=dict(
            description='The file path for FileHandler (e.g., app.log).'
        )
    )

    # * method: format_config
    def format_config(self) -> Dict[str, Any]:
        '''
        Format the handler configuration into a dictionary.

        :return: The formatted handler configuration.
        :rtype: Dict[str, Any]
        '''

        # Create the handler configuration.
        config = {
            'class': f'{self.module_path}.{self.class_name}',
            'level': self.level,
            'formatter': self.formatter
        }

        # Add optional attributes if they are set.
        if self.stream:
            config['stream'] = self.stream
        if self.filename:
            config['filename'] = self.filename

        # Return the handler configuration.
        return config


# ** model: logger
class Logger(DomainObject):
    '''
    An entity representing a logger configuration.
    '''

    # * attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The unique identifier of the logger.'
        )
    )

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the logger.'
        )
    )

    # * attribute: description
    description = StringType(
        metadata=dict(
            description='The description of the logger.'
        )
    )

    # * attribute: level
    level = StringType(
        required=True,
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
        metadata=dict(
            description='The logging level for the logger (e.g., DEBUG, WARNING).'
        )
    )

    # * attribute: handlers
    handlers = ListType(
        StringType(),
        default=[],
        metadata=dict(
            description='List of handler ids for the logger.'
        )
    )

    # * attribute: propagate
    propagate = BooleanType(
        default=False,
        metadata=dict(
            description='Whether to propagate messages to parent loggers.'
        )
    )

    # * attribute: is_root
    is_root = BooleanType(
        default=False,
        metadata=dict(
            description='Whether this is the root logger.'
        )
    )

    # * method: format_config
    def format_config(self) -> Dict[str, Any]:
        '''
        Format the logger configuration into a dictionary.

        :return: The formatted logger configuration.
        :rtype: Dict[str, Any]
        '''

        # Return the logger configuration.
        return {
            'level': self.level,
            'handlers': self.handlers,
            'propagate': self.propagate
        }



"""Tiferet Domain DI"""

# *** imports

# ** core
from importlib import import_module
from typing import Dict

# ** app
from .settings import (
    DomainObject,
    StringType,
    ListType,
    DictType,
    ModelType,
)

# *** models

# ** model: flagged_dependency
class FlaggedDependency(DomainObject):
    '''
    A flagged container dependency object.
    '''

    # * attribute: module_path
    module_path = StringType(
        required=True,
        metadata=dict(
            description='The module path.'
        ),
    )

    # * attribute: class_name
    class_name = StringType(
        required=True,
        metadata=dict(
            description='The class name.'
        ),
    )

    # * attribute: flag
    flag = StringType(
        required=True,
        metadata=dict(
            description='The flag for the container dependency.'
        ),
    )

    # * attribute: parameters
    parameters = DictType(
        StringType,
        default={},
        metadata=dict(
            description='The container dependency parameters.'
        ),
    )


# ** model: service_configuration
class ServiceConfiguration(DomainObject):
    '''
    A service configuration that defines dependency injection behavior.
    '''

    # * attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The unique identifier for the service configuration.'
        ),
    )

    # * attribute: name
    name = StringType(
        metadata=dict(
            description='The name of the service configuration.'
        ),
    )

    # * attribute: module_path
    module_path = StringType(
        metadata=dict(
            description='The default module path for the dependency class.'
        ),
    )

    # * attribute: class_name
    class_name = StringType(
        metadata=dict(
            description='The default class name for the dependency class.'
        ),
    )

    # * attribute: parameters
    parameters = DictType(
        StringType,
        default={},
        metadata=dict(
            description='The default configuration parameters.'
        ),
    )

    # * attribute: dependencies
    dependencies = ListType(
        ModelType(FlaggedDependency),
        default=[],
        metadata=dict(
            description='The flag-specific implementation overrides.'
        ),
    )

    # * method: get_dependency
    def get_dependency(self, *flags) -> FlaggedDependency:
        '''
        Gets a flagged dependency by flag.

        :param flags: The flags to match against flagged dependencies.
        :type flags: Tuple[str, ...]
        :return: The first flagged dependency matching any provided flag, or None.
        :rtype: FlaggedDependency
        '''

        # Return the first dependency that matches any of the provided flags.
        # Input flags are assumed ordinal in priority, so the first match is returned.
        for flag in flags:
            match = next(
                (dependency for dependency in self.dependencies if dependency.flag == flag),
                None
            )
            if match:
                return match

        # Return None if no dependency matches the flags.
        return None

    # * method: get_service_type
    def get_service_type(self, *flags) -> type:
        '''
        Gets the service type based on the provided flags.

        Checks flagged dependencies first (in flag priority order), then
        falls back to the configuration\'s default module_path/class_name.

        :param flags: The flags for the flagged dependency.
        :type flags: Tuple[str, ...]
        :return: The type of the service configuration.
        :rtype: type
        '''

        # Check the flagged dependencies for the type first.
        for flag in flags:
            dependency = self.get_dependency(flag)
            if dependency:
                return getattr(import_module(dependency.module_path), dependency.class_name)

        # Otherwise defer to an available default type.
        if self.module_path and self.class_name:
            return getattr(import_module(self.module_path), self.class_name)

        # Return None if no type is found.
        return None



"""Tiferet DI YAML Repository"""

# *** imports

# ** core
from typing import Any, Dict, List, Tuple

# ** app
from ..interfaces import DIService
from ..mappers import (
    TransferObject,
    ServiceConfigurationAggregate,
    ServiceConfigurationYamlObject,
)
from ..utils import Yaml

# *** repos

# ** repo: di_yaml_repository
class DIYamlRepository(DIService):
    '''
    The DI YAML repository.
    '''

    # * attribute: yaml_file
    yaml_file: str

    # * attribute: encoding
    encoding: str

    # * attribute: default_role
    default_role: str

    # * init
    def __init__(self, di_yaml_file: str, encoding: str = 'utf-8') -> None:
        '''
        Initialize the DI YAML repository.

        :param di_yaml_file: The YAML configuration file path.
        :type di_yaml_file: str
        :param encoding: The file encoding (default is 'utf-8').
        :type encoding: str
        '''

        # Set the repository attributes.
        self.yaml_file = di_yaml_file
        self.encoding = encoding
        self.default_role = 'to_data.yaml'

    # * method: configuration_exists
    def configuration_exists(self, id: str) -> bool:
        '''
        Check if a service configuration exists by ID.

        :param id: The service configuration identifier.
        :type id: str
        :return: True if the service configuration exists, otherwise False.
        :rtype: bool
        '''

        # Load the services mapping from the configuration file.
        services_data = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load(
            start_node=lambda data: data.get('services', {})
        )

        # Return whether the configuration id exists in the mapping.
        return id in services_data

    # * method: get_configuration
    def get_configuration(self, configuration_id: str, flag: str = None) -> ServiceConfigurationAggregate | None:
        '''
        Retrieve a service configuration by ID, optionally filtered by flag.

        :param configuration_id: The service configuration identifier.
        :type configuration_id: str
        :param flag: Optional flag to filter the configuration.
        :type flag: str
        :return: The service configuration aggregate or None if not found.
        :rtype: ServiceConfigurationAggregate | None
        '''

        # Load the specific service configuration data from the configuration file.
        config_data = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load(
            start_node=lambda data: data.get('services', {}).get(configuration_id)
        )

        # If no data is found, return None.
        if not config_data:
            return None

        # Map the data to a ServiceConfigurationAggregate.
        configuration = TransferObject.from_data(
            ServiceConfigurationYamlObject,
            id=configuration_id,
            **config_data,
        ).map()

        # If a flag is provided, filter the configuration by flag.
        if flag:
            dependency = configuration.get_dependency(flag)
            if dependency:
                configuration.module_path = dependency.module_path
                configuration.class_name = dependency.class_name
                if dependency.parameters:
                    merged = dict(configuration.parameters or {})
                    merged.update(dependency.parameters)
                    configuration.parameters = merged

        # Return the configuration.
        return configuration

    # * method: list_all
    def list_all(self) -> Tuple[List[ServiceConfigurationAggregate], Dict[str, str]]:
        '''
        List all service configurations and constants.

        :return: A tuple containing a list of service configuration aggregates and a dictionary of constants.
        :rtype: Tuple[List[ServiceConfigurationAggregate], Dict[str, str]]
        '''

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load()

        # Map each service entry to a ServiceConfigurationAggregate.
        configurations = [
            TransferObject.from_data(
                ServiceConfigurationYamlObject,
                id=config_id,
                **config_data,
            ).map()
            for config_id, config_data in full_data.get('services', {}).items()
        ]

        # Get the constants dictionary.
        constants = full_data.get('const', {})

        # Return the configurations and constants.
        return configurations, constants

    # * method: save_configuration
    def save_configuration(self, configuration: ServiceConfigurationAggregate) -> None:
        '''
        Save or update a service configuration.

        :param configuration: The service configuration aggregate to save.
        :type configuration: ServiceConfigurationAggregate
        :return: None
        :rtype: None
        '''

        # Convert the service configuration model to configuration data.
        config_data = ServiceConfigurationYamlObject.from_model(configuration)

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load()

        # Update or insert the service configuration entry.
        full_data.setdefault('services', {})[configuration.id] = config_data.to_primitive(self.default_role)

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding,
        ).save(data=full_data)

    # * method: delete_configuration
    def delete_configuration(self, configuration_id: str) -> None:
        '''
        Delete a service configuration by ID. This operation is idempotent.

        :param configuration_id: The service configuration identifier.
        :type configuration_id: str
        :return: None
        :rtype: None
        '''

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load()

        # Remove the service configuration entry if it exists (idempotent).
        full_data.get('services', {}).pop(configuration_id, None)

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding,
        ).save(data=full_data)

    # * method: save_constants
    def save_constants(self, constants: Dict[str, Any] = {}) -> None:
        '''
        Save or update constants.

        :param constants: The constants to save.
        :type constants: Dict[str, Any]
        :return: None
        :rtype: None
        '''

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding,
        ).load()

        # Merge existing constants with new ones.
        existing_constants = full_data.get('const', {})
        existing_constants.update(constants)

        # Remove any constants with None values.
        full_data['const'] = {k: v for k, v in existing_constants.items() if v is not None}

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding,
        ).save(data=full_data)



"""Tiferet Error Mappers"""

# *** imports

# ** core
from typing import Dict, Any

# ** app
from ..domain import (
    Error,
    ErrorMessage,
    ListType,
    ModelType,
)
from .settings import (
    Aggregate,
    TransferObject,
)

# *** mappers

# ** mapper: error_aggregate
class ErrorAggregate(Error, Aggregate):
    '''
    An aggregate representation of an error.
    '''

    # * method: new
    @staticmethod
    def new(
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> 'ErrorAggregate':
        '''
        Initializes a new error aggregate.

        :param validate: True to validate the aggregate object.
        :type validate: bool
        :param strict: True to enforce strict mode for the aggregate object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new error aggregate.
        :rtype: ErrorAggregate
        '''

        # Create a new error aggregate from the provided data.
        return Aggregate.new(
            ErrorAggregate,
            validate=validate,
            strict=strict,
            **kwargs
        )

    # * method: rename
    def rename(self, new_name: str) -> None:
        '''
        Renames the error.

        :param new_name: The new name for the error.
        :type new_name: str
        :return: None
        :rtype: None
        '''

        # Update the name.
        self.name = new_name

        # Perform final aggregate validation.
        self.validate()

    # * method: set_message
    def set_message(self, lang: str, text: str) -> None:
        '''
        Sets the error message text for the specified language.

        :param lang: The language of the error message text.
        :type lang: str
        :param text: The error message text.
        :type text: str
        :return: None
        :rtype: None
        '''

        # Check if the message already exists for the language.
        for msg in self.message:
            if msg.lang == lang:
                msg.text = text
                return

        # If not, create a new ErrorMessage object and add it to the message list.
        from ..domain import DomainObject
        self.message.append(
            DomainObject.new(
                ErrorMessage,
                lang=lang,
                text=text
            )
        )

    # * method: remove_message
    def remove_message(self, lang: str) -> None:
        '''
        Removes the error message for the specified language.

        :param lang: The language of the error message to remove.
        :type lang: str
        :return: None
        :rtype: None
        '''

        # Filter out the message with the specified language.
        self.message = [msg for msg in self.message if msg.lang != lang]

# ** mapper: error_message_yaml_object
class ErrorMessageYamlObject(ErrorMessage, TransferObject):
    '''
    A YAML data representation of an error message object.
    '''

    class Options():
        '''
        The options for the error message data.
        '''

        serialize_when_none = False
        roles = {
            'to_model': TransferObject.deny(),
            'to_data.yaml': TransferObject.deny(),
        }

    # * method: map
    def map(self, **kwargs) -> ErrorMessage:
        '''
        Maps the error message data to an error message object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new error message object.
        :rtype: ErrorMessage
        '''

        # Map to the error message object.
        return super().map(
            ErrorMessage,
            **self.to_primitive('to_model'),
            **kwargs
        )

    # * method: from_model
    @staticmethod
    def from_model(error_message: ErrorMessage, **kwargs) -> 'ErrorMessageYamlObject':
        '''
        Creates an ErrorMessageYamlObject from an ErrorMessage model.

        :param error_message: The error message model.
        :type error_message: ErrorMessage
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new ErrorMessageYamlObject.
        :rtype: ErrorMessageYamlObject
        '''

        # Create a new ErrorMessageYamlObject from the model.
        return TransferObject.from_model(
            ErrorMessageYamlObject,
            error_message,
            **kwargs,
        )


# ** mapper: error_yaml_object
class ErrorYamlObject(Error, TransferObject):
    '''
    A YAML data representation of an error object.
    '''

    class Options():
        '''
        The options for the error data.
        '''

        serialize_when_none = False
        roles = {
            'to_model': TransferObject.deny('message'),
            'to_data.yaml': TransferObject.deny('id'),
        }

    # * attribute: message
    message = ListType(
        ModelType(ErrorMessageYamlObject),
        required=True,
        metadata=dict(
            description='The error messages.'
        )
    )

    # * method: map
    def map(self, **kwargs) -> ErrorAggregate:
        '''
        Maps the error data to an error aggregate.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new error aggregate.
        :rtype: ErrorAggregate
        '''

        # Map the error data.
        return super().map(
            ErrorAggregate,
            message=[msg.map() for msg in self.message],
            **self.to_primitive('to_model'),
            **kwargs
        )

    # * method: from_model
    @staticmethod
    def from_model(error: Error, **kwargs) -> 'ErrorYamlObject':
        '''
        Creates an ErrorYamlObject from an Error model.

        :param error: The error model.
        :type error: Error
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new ErrorYamlObject.
        :rtype: ErrorYamlObject
        '''

        # Create a new ErrorYamlObject from the model, converting
        # the message list into ErrorMessageYamlObject instances.
        return TransferObject.from_model(
            ErrorYamlObject,
            error,
            message=[
                TransferObject.from_model(ErrorMessageYamlObject, msg)
                for msg in error.message
            ],
            **kwargs,
        )



# *** imports

# ** core
from typing import Any, Callable, Dict

# ** app
from ..assets import (
    TiferetError,
    TiferetAPIError,
    ERROR_NOT_FOUND_ID,
    DEFAULT_ERRORS
)
from ..domain import Error
from ..events import DomainEvent

# *** contexts

# ** context: error_context
class ErrorContext(object):
    '''
    The error context object.
    '''

    # * attribute: get_error_handler
    get_error_handler: Callable

    # * init
    def __init__(self, get_error_evt: DomainEvent):
        '''
        Initialize the error context.

        :param get_error_evt: The event to get an error by id.
        :type get_error_evt: DomainEvent
        '''

        # Assign the attributes.
        self.get_error_handler = get_error_evt.execute
    
    # * method: get_error_by_code
    def get_error_by_code(self, error_code: str) -> Error:
        '''
        Get an error by its code.

        :param error_code: The error code to retrieve.
        :type error_code: str
        :return: The error object.
        :rtype: Error
        '''

        # Try to retrieve the error by its code.
        try:
            return self.get_error_handler(error_code, include_defaults=True)
        
        # If the error is not found, raise the "error not found" error.
        except TiferetError:
            
            # Retrieve and raise the "error not found" error to use its details.
            error: Error = Error.new(**DEFAULT_ERRORS.get(ERROR_NOT_FOUND_ID))
            raise TiferetAPIError(
                **error.format_response(),
                id=error_code
            )
        
        # Return the retrieved error.
        return error

    # * method: handle_error
    def handle_error(self, exception: TiferetError, lang: str = 'en_US') -> Dict[str, Any]:
        '''
        Format and return the structured error response dictionary.
        Does not raise — raising is now handled by the calling context.

        :param exception: The exception to handle.
        :type exception: TiferetError
        :param lang: The language to use for the error message.
        :type lang: str
        :return: The formatted error response dictionary.
        :rtype: Dict[str, Any]
        '''

        # Raise the exception if it is not a TiferetError.
        if not isinstance(exception, TiferetError):
            raise exception

        # Get the error by its code from the error service.
        error = self.get_error_by_code(exception.error_code)

        # Return the formatted response dictionary (no raise).
        return error.format_response(lang=lang, **exception.kwargs)



# *** imports

# ** core
import sys
import argparse
from typing import Dict, Any, Callable

# ** app
from .app import *
from ..domain import CliCommand

# *** contexts

# ** context: cli_context
class CliContext(AppInterfaceContext):
    '''
    The CLI context is used to manage the command line interface of the application.
    It provides methods to handle command line arguments, commands, and their execution.
    '''

    # * attribute: list_cli_commands_handler
    list_cli_commands_handler: Callable

    # * attribute: get_parent_args_handler
    get_parent_args_handler: Callable

    # * init
    def __init__(self,
                 interface_id: str,
                 features: FeatureContext,
                 errors: ErrorContext,
                 logging: LoggingContext,
                 list_commands_evt: DomainEvent,
                 get_parent_args_evt: DomainEvent):
        '''
        Initialize the CLI context with command handlers.

        :param interface_id: The unique identifier for the CLI interface.
        :type interface_id: str
        :param features: The feature context.
        :type features: FeatureContext
        :param errors: The error context.
        :type errors: ErrorContext
        :param logging: The logging context.
        :type logging: LoggingContext
        :param list_commands_evt: Event to list all CLI commands.
        :type list_commands_evt: DomainEvent
        :param get_parent_args_evt: Event to get parent arguments.
        :type get_parent_args_evt: DomainEvent
        '''

        # Set the command handlers.
        self.list_cli_commands_handler = list_commands_evt.execute
        self.get_parent_args_handler = get_parent_args_evt.execute

        # Initialize the base class with the interface ID, features, and errors.
        super().__init__(
            interface_id=interface_id,
            features=features,
            errors=errors,
            logging=logging
        )

    # * method: get_commands
    def get_commands(self) -> Dict[str, CliCommand]:
        '''
        Get all commands available in the CLI mapped by their group keys.

        :return: A dictionary of CLI commands mapped by their group keys.
        :rtype: Dict[str, CliCommand]
        '''

        # Retrieve the commands via the list commands handler.
        cli_commands = self.list_cli_commands_handler()

        # Create a map of commands by their group keys.
        command_map = {}
        for command in cli_commands:
            if command.group_key not in command_map:
                command_map[command.group_key] = [command]
            else:
                command_map[command.group_key].append(command)

        # Return the command map.
        return command_map

    # * method: parse_arguments
    def parse_arguments(self, cli_commands: Dict[str, CliCommand]) -> Dict[str, Any]:
        '''
        Parse the command line arguments for a list of CLI commands.

        :param cli_commands: The CLI commands to parse arguments for.
        :type cli_commands: Dict[str, CliCommand]
        :return: A dictionary of parsed arguments.
        :rtype: Dict[str, Any]
        '''

        # Retrieve the parent arguments via the get parent args handler.
        parent_arguments = self.get_parent_args_handler()

        # Create an argument parser.
        parser = argparse.ArgumentParser()

        # Add command subparsers for the command group.
        group_subparsers = parser.add_subparsers(dest='group')

        # Loop through the command map and create a parser for each command.
        for group_key in cli_commands:
            group_subparser = group_subparsers.add_parser(
                group_key,
                help=f'Commands for the {group_key} group.'
            )
            cli_group_commands = cli_commands[group_key]
            cmd_subparsers = group_subparser.add_subparsers(dest='command')

            for cli_command in cli_group_commands:
                cli_command_parser = cmd_subparsers.add_parser(
                    cli_command.key,
                    help=cli_command.description
                )

                for argument in cli_command.arguments:
                    args = dict(
                        help=argument.description,
                        type=argument.get_type(),
                        default=argument.default,
                        nargs=argument.nargs,
                        choices=argument.choices,
                        action=argument.action
                    )
                    if argument.required is not None:
                        args['required'] = argument.required
                    cli_command_parser.add_argument(*argument.name_or_flags, **args)

                for argument in parent_arguments:
                    if not cli_command.has_argument(argument.name_or_flags):
                        cli_command_parser.add_argument(
                            *argument.name_or_flags,
                            help=argument.description,
                            type=argument.get_type(),
                            default=argument.default,
                            required=argument.required,
                            nargs=argument.nargs,
                            choices=argument.choices,
                            action=argument.action
                        )

        # Return the parsed arguments as a dictionary.
        return vars(parser.parse_args())

    # * method: parse_request
    def parse_request(self) -> RequestContext:
        '''
        Parse the command line arguments and return a Request object.

        :return: The parsed request context.
        :rtype: RequestContext
        '''

        # Retrieve the command map via get_commands.
        cli_commands = self.get_commands()

        # Parse the command line arguments.
        data = self.parse_arguments(cli_commands)

        # Fashion the feature id from the command group and key.
        feature_id = '{}.{}'.format(
            data.get('group').replace('-', '_'),
            data.get('command').replace('-', '_')
        )

        # Create a RequestContext object with the parsed data and headers.
        return super().parse_request(
            headers=dict(
                command_group=data.get('group'),
                command_key=data.get('command'),
            ),
            data=data,
            feature_id=feature_id,
        )

    # * method: run
    def run(self) -> Any:
        '''
        Run the CLI context by parsing the request and executing the command.

        :return: The result of the command execution.
        :rtype: Any
        '''

        # Create a logger for the CLI context.
        logger = self.logging.build_logger()

        # Attempt to parse the command line request.
        try:
            logger.debug('Parsing CLI request...')
            cli_request = self.parse_request()

        # Handle any exceptions that may occur during request parsing.
        except Exception as e:
            logger.error(f'Error parsing CLI request: {e}')
            print(e, file=sys.stderr)
            sys.exit(2)

        # Attempt to execute the feature for the parsed CLI request.
        try:
            logger.info(f'Executing feature for CLI request: {cli_request.feature_id}')
            logger.debug(f'CLI request data: {cli_request.data}')
            self.execute_feature(
                feature_id=cli_request.feature_id,
                request=cli_request,
                logger=logger,
            )
        
        # Handle any TiferetError exceptions that may occur during feature execution.
        except TiferetError as e:
            logger.error(f'Error executing CLI feature {cli_request.feature_id}: {e}')
            try:
                self.handle_error(e)
            except TiferetAPIError as api_error:
                print(api_error, file=sys.stderr)
            sys.exit(1)
        
        # Return the result of the command execution.
        logger.debug('CLI request executed successfully.')
        print(cli_request.handle_response())



"""Tiferet Feature Events"""

# *** imports

# ** core
from typing import List, Any

# ** app
from ..domain import Feature
from ..interfaces import FeatureService
from ..mappers import FeatureAggregate
from .settings import DomainEvent, a

# *** events

# ** event: add_feature
class AddFeature(DomainEvent):
    '''
    Event to add a new feature configuration.
    '''

    # * attribute: feature_service
    feature_service: FeatureService

    # * init
    def __init__(self, feature_service: FeatureService):
        '''
        Initialize the AddFeature event.

        :param feature_service: The feature service to use for managing feature configurations.
        :type feature_service: FeatureService
        '''

        # Set the feature service dependency.
        self.feature_service = feature_service

    # * method: execute
    @DomainEvent.parameters_required(['name', 'group_id'])
    def execute(
            self,
            name: str,
            group_id: str,
            feature_key: str | None = None,
            id: str | None = None,
            description: str | None = None,
            steps: list | None = None,
            log_params: dict | None = None,
            **kwargs,
        ) -> Feature:
        '''
        Add a new feature.

        :param name: Required feature name.
        :type name: str
        :param group_id: Required group identifier.
        :type group_id: str
        :param feature_key: Optional explicit key (defaults to snake_case of name).
        :type feature_key: str | None
        :param id: Optional explicit full ID (defaults to f'{group_id}.{feature_key}').
        :type id: str | None
        :param description: Optional description (defaults to name).
        :type description: str | None
        :param steps: Optional list of initial feature steps.
        :type steps: list | None
        :param log_params: Optional logging parameters.
        :type log_params: dict | None
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The created Feature model.
        :rtype: Feature
        '''

        # Create feature using the aggregate factory.
        feature = FeatureAggregate.new(
            name=name,
            group_id=group_id,
            feature_key=feature_key,
            id=id,
            description=description,
            steps=steps or [],
            log_params=log_params or {},
            **kwargs,
        )

        # Check for duplicate feature identifier.
        self.verify(
            expression=not self.feature_service.exists(feature.id),
            error_code=a.const.FEATURE_ALREADY_EXISTS_ID,
            message=f'Feature with ID {feature.id} already exists.',
            id=feature.id,
        )

        # Persist the new feature.
        self.feature_service.save(feature)

        # Return the created feature.
        return feature

# ** event: get_feature
class GetFeature(DomainEvent):
    '''
    Event to retrieve a feature by its identifier.
    '''

    # * attribute: feature_service
    feature_service: FeatureService

    # * init
    def __init__(self, feature_service: FeatureService):
        '''
        Initialize the GetFeature event.

        :param feature_service: The feature service to use for retrieving features.
        :type feature_service: FeatureService
        '''

        # Set the feature service dependency.
        self.feature_service = feature_service

    # * method: execute
    @DomainEvent.parameters_required(['id'])
    def execute(self, id: str, **kwargs) -> Feature:
        '''
        Retrieve a feature by ID.

        :param id: The feature identifier.
        :type id: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The retrieved feature.
        :rtype: Feature
        '''

        # Retrieve the feature from the feature service.
        feature = self.feature_service.get(id)

        # Verify that the feature exists; raise FEATURE_NOT_FOUND if it does not.
        self.verify(
            expression=feature is not None,
            error_code=a.const.FEATURE_NOT_FOUND_ID,
            feature_id=id,
        )

        # Return the retrieved feature.
        return feature

# ** event: list_features
class ListFeatures(DomainEvent):
    '''
    Event to list feature configurations.
    '''

    # * attribute: feature_service
    feature_service: FeatureService

    # * init
    def __init__(self, feature_service: FeatureService):
        '''
        Initialize the ListFeatures event.

        :param feature_service: The feature service to use for listing features.
        :type feature_service: FeatureService
        '''

        # Set the feature service dependency.
        self.feature_service = feature_service

    # * method: execute
    def execute(self, group_id: str | None = None, **kwargs) -> List[Feature]:
        '''
        List features, optionally filtered by group_id.

        :param group_id: Optional group identifier to filter results.
        :type group_id: str | None
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: List of Feature models.
        :rtype: List[Feature]
        '''

        # Delegate to the feature service.
        return self.feature_service.list(group_id=group_id)

# ** event: remove_feature
class RemoveFeature(DomainEvent):
    '''
    Event to remove an entire feature configuration by ID (idempotent).

    This event delegates deletion semantics to the underlying
    ``FeatureService.delete`` implementation, which is expected to behave
    idempotently when the feature does not exist.
    '''

    # * attribute: feature_service
    feature_service: FeatureService

    # * init
    def __init__(self, feature_service: FeatureService):
        '''
        Initialize the RemoveFeature event.

        :param feature_service: The feature service to use.
        :type feature_service: FeatureService
        '''

        # Set the feature service dependency.
        self.feature_service = feature_service

    # * method: execute
    @DomainEvent.parameters_required(['id'])
    def execute(self, id: str, **kwargs) -> str:
        '''
        Remove a feature by ID.

        :param id: The feature ID.
        :type id: str
        :param kwargs: Additional keyword arguments (unused).
        :type kwargs: dict
        :return: The removed feature ID.
        :rtype: str
        '''

        # Delete the feature using the feature service.
        # and treated as a successful no-op when the feature does not exist.
        self.feature_service.delete(id)

        # Return the feature identifier.
        return id

# ** event: update_feature
class UpdateFeature(DomainEvent):
    '''
    Event to update basic metadata of an existing feature.

    Supports updating the ``name`` or ``description`` attributes using the
    Feature model helpers.
    '''

    # * attribute: feature_service
    feature_service: FeatureService

    # * init
    def __init__(self, feature_service: FeatureService) -> None:
        '''
        Initialize the UpdateFeature event.

        :param feature_service: The feature service used to retrieve and
            persist features.
        :type feature_service: FeatureService
        '''

        # Set the feature service dependency.
        self.feature_service = feature_service

    # * method: execute
    @DomainEvent.parameters_required(['id', 'attribute'])
    def execute(
            self,
            id: str,
            attribute: str,
            value: Any,
            **kwargs,
        ) -> Feature:
        '''
        Update a feature's ``name`` or ``description`` attribute.

        :param id: The identifier of the feature to update.
        :type id: str
        :param attribute: The attribute to update (``"name"`` or
            ``"description"``).
        :type attribute: str
        :param value: The new value for the attribute.
        :type value: Any
        :param kwargs: Additional keyword arguments (unused).
        :type kwargs: dict
        :return: The updated Feature instance.
        :rtype: Feature
        '''

        # Validate that the attribute is supported.
        self.verify(
            expression=attribute in ('name', 'description'),
            error_code=a.const.INVALID_FEATURE_ATTRIBUTE_ID,
            message=f'Invalid feature attribute: {attribute}',
            attribute=attribute,
        )

        # When updating the name, ensure a non-empty value is provided.
        if attribute == 'name':
            self.verify(
                expression=isinstance(value, str) and bool(value.strip()),
                error_code=a.const.FEATURE_NAME_REQUIRED_ID,
                message='A feature name is required when updating the name attribute.',
            )

        # Retrieve the feature from the feature service.
        feature = self.feature_service.get(id)

        # Verify that the feature exists.
        self.verify(
            expression=feature is not None,
            error_code=a.const.FEATURE_NOT_FOUND_ID,
            feature_id=id,
        )

        # Apply the requested update using model helpers.
        if attribute == 'name':
            feature.rename(value)
        elif attribute == 'description':
            feature.set_description(value)

        # Persist the updated feature.
        self.feature_service.save(feature)

        # Return the updated feature.
        return feature

# ** event: add_feature_step
class AddFeatureStep(DomainEvent):
    '''
    Event to add a step to an existing feature.
    '''

    # * attribute: feature_service
    feature_service: FeatureService

    # * init
    def __init__(self, feature_service: FeatureService):
        '''
        Initialize the AddFeatureStep event.

        :param feature_service: The feature service to use.
        :type feature_service: FeatureService
        '''

        # Set the feature service dependency.
        self.feature_service = feature_service

    # * method: execute
    @DomainEvent.parameters_required(['id', 'name', 'service_id'])
    def execute(
            self,
            id: str,
            name: str,
            service_id: str,
            parameters: dict | None = None,
            data_key: str | None = None,
            pass_on_error: bool = False,
            position: int | None = None,
            **kwargs,
        ) -> str:
        '''
        Add a step to an existing feature.

        :param id: The feature ID.
        :type id: str
        :param name: The step name.
        :type name: str
        :param service_id: The service configuration ID for the step.
        :type service_id: str
        :param parameters: Optional step parameters.
        :type parameters: dict | None
        :param data_key: Optional result data key.
        :type data_key: str | None
        :param pass_on_error: Whether to pass on errors from this step.
        :type pass_on_error: bool
        :param position: Insertion position (None to append).
        :type position: int | None
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The feature ID.
        :rtype: str
        '''

        # Retrieve the feature from the feature service.
        feature = self.feature_service.get(id)
        self.verify(
            expression=feature is not None,
            error_code=a.const.FEATURE_NOT_FOUND_ID,
            message=f'Feature not found: {id}',
            feature_id=id,
        )

        # Add the step using the Feature model helper.
        feature.add_step(
            name=name,
            service_id=service_id,
            parameters=parameters or {},
            data_key=data_key,
            pass_on_error=pass_on_error,
            position=position,
        )

        # Persist the updated feature.
        self.feature_service.save(feature)

        # Return the feature identifier.
        return id

# ** event: update_feature_step
class UpdateFeatureStep(DomainEvent):
    '''
    Event to update an existing feature step within a feature workflow.

    This event supports updating the following attributes on a
    ``FeatureEvent`` instance: ``name``, ``service_id``, ``data_key``,
    ``pass_on_error``, and ``parameters``.
    '''

    # * attribute: feature_service
    feature_service: FeatureService

    # * init
    def __init__(self, feature_service: FeatureService) -> None:
        '''
        Initialize the UpdateFeatureStep event.

        :param feature_service: The feature service used to retrieve and
            persist features.
        :type feature_service: FeatureService
        '''

        # Set the feature service dependency.
        self.feature_service = feature_service

    # * method: execute
    @DomainEvent.parameters_required(['id', 'position', 'attribute'])
    def execute(
            self,
            id: str,
            position: int,
            attribute: str,
            value: Any | None = None,
            **kwargs,
        ) -> str:
        '''
        Update an attribute on a feature step at the given position.

        :param id: The identifier of the feature whose step will be
            updated.
        :type id: str
        :param position: The zero-based index of the step within the
            feature's step list.
        :type position: int
        :param attribute: The attribute to update. Supported values are
            ``"name"``, ``"service_id"``, ``"data_key"``,
            ``"pass_on_error"``, and ``"parameters"``.
        :type attribute: str
        :param value: The new value for the attribute. For ``name`` and
            ``service_id`` this must be a non-empty value.
        :type value: Any | None
        :param kwargs: Additional keyword arguments (unused).
        :type kwargs: dict
        :return: The feature identifier.
        :rtype: str
        '''

        # Validate that the attribute name is supported.
        valid_attributes = {
            'name',
            'service_id',
            'data_key',
            'pass_on_error',
            'parameters',
        }
        self.verify(
            expression=attribute in valid_attributes,
            error_code=a.const.INVALID_FEATURE_COMMAND_ATTRIBUTE_ID,
            message=(
                'Invalid feature step attribute: {attribute}. '
                'Supported attributes are name, service_id, data_key, '
                'pass_on_error, and parameters.'
            ),
            attribute=attribute,
        )

        # For name and service_id, enforce a non-empty value.
        if attribute in {'name', 'service_id'}:
            self.verify(
                expression=(
                    value is not None
                    and (not isinstance(value, str) or bool(str(value).strip()))
                ),
                error_code=a.const.COMMAND_PARAMETER_REQUIRED_ID,
                message=(
                    f'The "value" parameter is required when updating the '
                    f'"{attribute}" attribute for the '
                    f'"{self.__class__.__name__}" command.'
                ),
                parameter='value',
                command=self.__class__.__name__,
            )

        # Retrieve the feature from the feature service.
        feature = self.feature_service.get(id)

        # Verify that the feature exists.
        self.verify(
            expression=feature is not None,
            error_code=a.const.FEATURE_NOT_FOUND_ID,
            feature_id=id,
        )

        # Retrieve the target step from the feature.
        step = feature.get_step(position)

        # Verify that the step exists at the given position.
        self.verify(
            expression=step is not None,
            error_code=a.const.FEATURE_COMMAND_NOT_FOUND_ID,
            message=(
                f'Feature step not found for feature {id} '
                f'at position {position}.'
            ),
            feature_id=id,
            position=position,
        )

        # Apply the attribute update using the FeatureEvent helper.
        step.set_attribute(attribute, value)

        # Persist the updated feature.
        self.feature_service.save(feature)

        # Return the feature identifier.
        return id

# ** event: remove_feature_step
class RemoveFeatureStep(DomainEvent):
    '''
    Event to remove a step from an existing feature by position.

    This event is idempotent: invalid positions result in silent success
    with no mutation to the feature's step list.
    '''

    # * attribute: feature_service
    feature_service: FeatureService

    # * init
    def __init__(self, feature_service: FeatureService) -> None:
        '''
        Initialize the RemoveFeatureStep event.

        :param feature_service: The feature service to use for retrieving and
            persisting features.
        :type feature_service: FeatureService
        '''

        # Set the feature service dependency.
        self.feature_service = feature_service

    # * method: execute
    @DomainEvent.parameters_required(['id', 'position'])
    def execute(
            self,
            id: str,
            position: int,
            **kwargs,
        ) -> str:
        '''
        Remove a step from the feature at the given position.

        :param id: The feature identifier.
        :type id: str
        :param position: The index of the step to remove.
        :type position: int
        :param kwargs: Additional keyword arguments (unused).
        :type kwargs: dict
        :return: The feature identifier.
        :rtype: str
        '''

        # Retrieve the feature from the feature service.
        feature = self.feature_service.get(id)

        # Verify that the feature exists.
        self.verify(
            expression=feature is not None,
            error_code=a.const.FEATURE_NOT_FOUND_ID,
            message=f'Feature not found: {id}',
            feature_id=id,
        )

        # Attempt safe removal of the step at the given position. The
        # underlying Feature.remove_step helper is idempotent and will
        # return None without raising if the position is invalid.
        feature.remove_step(position)

        # Persist the feature, even if no step was removed.
        self.feature_service.save(feature)

        # Return the feature identifier.
        return id

# ** event: reorder_feature_step
class ReorderFeatureStep(DomainEvent):
    '''
    Event to reorder an existing feature step within a feature workflow.

    This event delegates to the ``Feature.reorder_step`` model helper,
    which clamps the target position and behaves idempotently for invalid
    start positions.
    '''

    # * attribute: feature_service
    feature_service: FeatureService

    # * init
    def __init__(self, feature_service: FeatureService) -> None:
        '''
        Initialize the ReorderFeatureStep event.

        :param feature_service: The feature service used to retrieve and
            persist features.
        :type feature_service: FeatureService
        '''

        # Set the feature service dependency.
        self.feature_service = feature_service

    # * method: execute
    @DomainEvent.parameters_required(['id', 'start_position', 'end_position'])
    def execute(
            self,
            id: str,
            start_position: int,
            end_position: int,
            **kwargs,
        ) -> str:
        '''
        Reorder a feature step by moving it from ``start_position`` to
        ``end_position`` within the feature's step list.

        :param id: The identifier of the feature whose step will be
            reordered.
        :type id: str
        :param start_position: The current index of the step to move.
        :type start_position: int
        :param end_position: The desired new index for the step.
        :type end_position: int
        :param kwargs: Additional keyword arguments (unused).
        :type kwargs: dict
        :return: The feature identifier.
        :rtype: str
        '''

        # Retrieve the feature from the feature service.
        feature = self.feature_service.get(id)

        # Verify that the feature exists.
        self.verify(
            expression=feature is not None,
            error_code=a.const.FEATURE_NOT_FOUND_ID,
            message=f'Feature not found: {id}',
            feature_id=id,
        )

        # Delegate the reordering logic to the Feature model helper. This
        # method clamps ``end_position`` and is idempotent for invalid
        # ``start_position`` values.
        feature.reorder_step(start_position, end_position)

        # Persist the updated feature configuration.
        self.feature_service.save(feature)

        # Return the feature identifier.
        return id



"""Dependencies DI Service Provider"""

# *** imports

# ** core
from typing import Any, Dict

# ** infra
from dependencies import Injector
from dependencies.exceptions import DependencyError

# ** app
from ..events import RaiseError
from .settings import ServiceProvider


# *** classes

# ** class: dependencies_service_provider
class DependenciesServiceProvider(ServiceProvider):
    '''
    A service provider for managing app context dependencies via the
    dependencies library injector.
    '''

    # * attribute: services
    services: Dict[str, type]

    # * attribute: injector
    injector: type

    # * attribute: _name
    _name: str

    # * init
    def __init__(self, services: Dict[str, type] = None, name: str = 'TiferetInjector'):
        '''
        Initialize the dependencies service provider.

        :param services: Initial service ID-to-type mapping.
        :type services: Dict[str, type]
        :param name: Name used for the internal injector class.
        :type name: str
        '''

        # Store the name and services, then build the injector.
        self._name = name
        self.services = services or {}
        self.build_injector()

    # * method: new (static)
    @staticmethod
    def new(name: str, dependencies: Dict[str, type]) -> 'DependenciesServiceProvider':
        '''
        Create a new service provider instance with the given dependencies.

        :param name: The name of the service provider.
        :type name: str
        :param dependencies: The dependencies to include in the service provider.
        :type dependencies: Dict[str, type]
        :return: A new service provider instance.
        :rtype: DependenciesServiceProvider
        '''

        # Return a new provider instance pre-populated with the given dependencies.
        return DependenciesServiceProvider(services=dependencies, name=name)

    # * method: add_service
    def add_service(self, service_id: str, service_type: type):
        '''
        Add a service dependency to the service provider.

        :param service_id: The ID of the service to add.
        :type service_id: str
        :param service_type: The type of the service to add.
        :type service_type: type
        '''

        # Add the service to the services dictionary and rebuild the injector.
        self.services[service_id] = service_type
        self.build_injector()

    # * method: add_services
    def add_services(self, services: Dict[str, type]):
        '''
        Add multiple service dependencies to the service provider.

        :param services: A dictionary of service IDs and their corresponding types.
        :type services: Dict[str, type]
        '''

        # Merge the given services and rebuild the injector.
        self.services.update(services)
        self.build_injector()

    # * method: add_constants
    def add_constants(self, constants: Dict[str, Any]):
        '''
        Add constant dependencies to the service provider.

        :param constants: A dictionary of constant names and their corresponding values.
        :type constants: Dict[str, Any]
        '''

        # Merge the constants into the services dictionary and rebuild the injector.
        self.services.update(constants)
        self.build_injector()

    # * method: get_service
    def get_service(self, service_id: str) -> Any:
        '''
        Get a service dependency by its ID.

        :param service_id: The service ID.
        :type service_id: str
        :return: The resolved service dependency instance.
        :rtype: Any
        '''

        # If no services have been registered the injector is None — raise immediately.
        if self.injector is None:
            RaiseError.execute(
                'INVALID_DEPENDENCY_ERROR',
                f'Dependency {service_id} could not be resolved: no services are registered.',
                dependency_name=service_id,
                exception='No services registered in the service provider.',
            )

        # Attempt to retrieve the resolved service from the injector.
        try:
            return getattr(self.injector, service_id)

        # If the dependency does not exist or cannot be resolved, raise an error.
        except (DependencyError, AttributeError) as e:
            RaiseError.execute(
                'INVALID_DEPENDENCY_ERROR',
                f'Dependency {service_id} could not be resolved: {str(e)}',
                dependency_name=service_id,
                exception=str(e)
            )

    # * method: remove_service
    def remove_service(self, service_id: str):
        '''
        Remove a service dependency from the service provider.

        :param service_id: The service ID to remove.
        :type service_id: str
        '''

        # Remove the service from the services dictionary and rebuild the injector.
        if service_id in self.services:
            del self.services[service_id]
            self.build_injector()

    # * method: build_injector
    def build_injector(self):
        '''
        Rebuild the internal injector from the current services dictionary.
        The injector is set to None when no services are registered, as the
        dependencies library requires a non-empty scope.
        '''

        # Only build the injector when there are services to register.
        if self.services:
            self.injector = type(self._name, (Injector,), self.services)
        else:
            self.injector = None



# *** imports

# ** core
from typing import Callable, Any, List, Dict

# ** app
from .cache import CacheContext
from ..assets.constants import DEPENDENCY_TYPE_NOT_FOUND_ID
from ..di import ServiceProvider, DependenciesServiceProvider
from ..domain.di import ServiceConfiguration
from ..events import DomainEvent, RaiseError, ParseParameter


# *** contexts

# ** context: di_context
class DIContext(object):
    '''
    A context for managing dependency injection configuration and service provider lifecycle.
    '''

    # * attribute: cache
    cache: CacheContext

    # * attribute: list_all_configs_handler
    list_all_configs_handler: Callable

    # * init
    def __init__(self, di_list_all_configs_evt: DomainEvent, cache: CacheContext = None):
        '''
        Initialize the DI context.

        :param di_list_all_configs_evt: The event to list all service configurations and constants.
        :type di_list_all_configs_evt: DomainEvent
        :param cache: The cache context to use for caching service providers.
        :type cache: CacheContext
        '''

        # Assign the attributes.
        self.list_all_configs_handler = di_list_all_configs_evt.execute
        self.cache = cache if cache else CacheContext()

    # * method: create_cache_key
    def create_cache_key(self, flags: List[str] = None) -> str:
        '''
        Create a cache key for the service provider.

        :param flags: The feature or data flags to use.
        :type flags: List[str]
        :return: The cache key.
        :rtype: str
        '''

        # Create the cache key from the flags.
        return f"feature_services{'_' + '_'.join(flags) if flags else ''}"

    # * method: build_provider
    def build_provider(self, flags: List[str] = []) -> ServiceProvider:
        '''
        Build and cache a service provider for the given flags.

        :param flags: The feature or data flags to use.
        :type flags: List[str]
        :return: The service provider instance.
        :rtype: ServiceProvider
        '''

        # Create the cache key from the flags.
        cache_key = self.create_cache_key(flags)

        # Return the cached provider if available.
        cached_provider = self.cache.get(cache_key)
        if cached_provider:
            return cached_provider

        # Get all service configurations and constants from the DI service.
        configurations, constants = self.list_all_configs_handler()

        # Load and parse constants from configurations and the top-level constants dict.
        constants = self.load_constants(configurations, constants, flags)

        # Resolve service types from configurations.
        services = {}
        for config in configurations:

            # Get the dependency type based on the flags.
            dep_type = self.get_configuration_type(config, *flags)

            # If no type is found, raise an error.
            if not dep_type:
                RaiseError.execute(
                    DEPENDENCY_TYPE_NOT_FOUND_ID,
                    f'No dependency type found for service configuration {config.id} with flags {flags}.',
                    configuration_id=config.id,
                    flags=flags,
                )

            # Add the resolved type to the services dictionary.
            services[config.id] = dep_type

        # Merge services and constants and build the provider.
        provider = DependenciesServiceProvider(
            services={**services, **constants},
            name=cache_key,
        )

        # Cache and return the provider.
        self.cache.set(cache_key, provider)
        return provider

    # * method: get_dependency
    def get_dependency(self, configuration_id: str, flags: List[str] = []) -> Any:
        '''
        Get a resolved service by its configuration ID.

        :param configuration_id: The service configuration identifier.
        :type configuration_id: str
        :param flags: The feature or data flags to use.
        :type flags: List[str]
        :return: The resolved service instance.
        :rtype: Any
        '''

        # Build (or retrieve) the service provider for these flags.
        provider = self.build_provider(flags)

        # Return the resolved service from the provider.
        return provider.get_service(configuration_id)

    # * method: get_configuration_type
    def get_configuration_type(self, configuration: ServiceConfiguration, *flags) -> type:
        '''
        Resolve the implementation type for a service configuration.

        Delegates to ServiceConfiguration.get_service_type() which checks
        flagged dependencies first, then falls back to the default type.

        :param configuration: The service configuration to resolve.
        :type configuration: ServiceConfiguration
        :param flags: The flags for the flagged dependency.
        :type flags: Tuple[str, ...]
        :return: The resolved type, or None.
        :rtype: type
        '''

        # Delegate to the domain method for type resolution.
        return configuration.get_service_type(*flags)

    # * method: load_constants
    def load_constants(self,
            configurations: List[ServiceConfiguration] = [],
            constants: Dict[str, str] = {},
            flags: List[str] = [],
        ) -> Dict[str, str]:
        '''
        Build the constants dict by parsing top-level constants and per-configuration parameters.

        :param configurations: The list of service configurations.
        :type configurations: List[ServiceConfiguration]
        :param constants: The top-level constants dictionary.
        :type constants: Dict[str, str]
        :param flags: The feature or data flags to use.
        :type flags: List[str]
        :return: A dictionary of parsed constants.
        :rtype: Dict[str, str]
        '''

        # Parse the top-level constants.
        constants = {k: ParseParameter.execute(v) for k, v in constants.items()}

        # Merge in per-configuration parameters (flagged or default).
        for config in configurations:

            # Check for a flagged dependency matching any of the provided flags.
            dependency = config.get_dependency(*flags)

            # Use flagged parameters if available; otherwise use the default parameters.
            if dependency:
                constants.update({k: ParseParameter.execute(v) for k, v in dependency.parameters.items()})
            else:
                constants.update({k: ParseParameter.execute(v) for k, v in config.parameters.items()})

        # Return the merged constants dictionary.
        return constants



"""Tiferet Interfaces Configuration"""

# *** imports

# ** core
from abc import abstractmethod
from typing import Any, Callable

# ** app
from .settings import Service

# *** interfaces

# ** interface: configuration_service
class ConfigurationService(Service):
    '''
    Service contract for loading and persisting structured configuration data.
    '''

    # * method: load
    @abstractmethod
    def load(self,
            start_node: Callable[[Any], Any] = lambda x: x,
            data_factory: Callable[[Any], Any] = lambda x: x,
            **kwargs) -> Any:
        '''
        Load structured configuration data, optionally transforming it.

        :param start_node: Callable to select the starting node of the data.
        :type start_node: Callable[[Any], Any]
        :param data_factory: Callable to transform the loaded data.
        :type data_factory: Callable[[Any], Any]
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The loaded configuration data.
        :rtype: Any
        '''
        raise NotImplementedError()

    # * method: save
    @abstractmethod
    def save(self, data: Any, data_path: str | None = None, **kwargs) -> None:
        '''
        Persist structured configuration data.

        :param data: The configuration data to persist.
        :type data: Any
        :param data_path: Optional path within the data structure.
        :type data_path: str | None
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''
        raise NotImplementedError()



"""Tiferet Interfaces SQLite"""

# *** imports

# ** core
from abc import abstractmethod
from typing import Any, Callable, Iterable, Optional

# ** app
from .file import FileService

# *** interfaces

# ** interface: sqlite_service
class SqliteService(FileService):
    '''
    Service contract for SQLite database operations.
    '''

    # * method: execute
    @abstractmethod
    def execute(self, sql: str, parameters: Iterable[Any] = ()) -> Any:
        '''
        Execute a single SQL statement.

        :param sql: The SQL statement to execute.
        :type sql: str
        :param parameters: Parameters for the SQL statement.
        :type parameters: Iterable[Any]
        :return: The cursor result.
        :rtype: Any
        '''
        raise NotImplementedError()

    # * method: executemany
    @abstractmethod
    def executemany(self, sql: str, seq_of_parameters: Iterable[Iterable[Any]]) -> Any:
        '''
        Execute SQL repeatedly with parameter sequences.

        :param sql: The SQL statement to execute.
        :type sql: str
        :param seq_of_parameters: Sequence of parameter sets.
        :type seq_of_parameters: Iterable[Iterable[Any]]
        :return: The cursor result.
        :rtype: Any
        '''
        raise NotImplementedError()

    # * method: executescript
    @abstractmethod
    def executescript(self, sql_script: str) -> Any:
        '''
        Execute multiple SQL statements from a script.

        :param sql_script: The SQL script to execute.
        :type sql_script: str
        :return: The cursor result.
        :rtype: Any
        '''
        raise NotImplementedError()

    # * method: fetch_one
    @abstractmethod
    def fetch_one(self, query: str, parameters: Iterable[Any] = ()) -> tuple | None:
        '''
        Execute a query and fetch a single row.

        :param query: The SQL query to execute.
        :type query: str
        :param parameters: Parameters for the SQL query.
        :type parameters: Iterable[Any]
        :return: The first row as a tuple, or None if no rows.
        :rtype: tuple | None
        '''
        raise NotImplementedError()

    # * method: fetch_all
    @abstractmethod
    def fetch_all(self, query: str, parameters: Iterable[Any] = ()) -> list[tuple]:
        '''
        Execute a query and fetch all rows.

        :param query: The SQL query to execute.
        :type query: str
        :param parameters: Parameters for the SQL query.
        :type parameters: Iterable[Any]
        :return: All rows as a list of tuples.
        :rtype: list[tuple]
        '''
        raise NotImplementedError()

    # * method: commit
    @abstractmethod
    def commit(self) -> None:
        '''
        Commit current transaction.
        '''
        raise NotImplementedError()

    # * method: rollback
    @abstractmethod
    def rollback(self) -> None:
        '''
        Roll back current transaction.
        '''
        raise NotImplementedError()

    # * method: backup
    @abstractmethod
    def backup(self,
            target_path: str,
            pages: int = -1,
            progress: Optional[Callable[[int, int, int], None]] = None,
        ) -> None:
        '''
        Backup database to a target file path.

        :param target_path: The file path for the backup database.
        :type target_path: str
        :param pages: Number of pages to copy at a time (-1 for all).
        :type pages: int
        :param progress: Optional progress callback(status, remaining, total).
        :type progress: Optional[Callable[[int, int, int], None]]
        '''
        raise NotImplementedError()



"""Tiferet Logging YAML Repository"""

# *** imports

# ** core
from typing import (
    Tuple,
    List
)

# ** app
from ..interfaces import LoggingService
from ..mappers import (
    LoggingSettingsYamlObject,
    FormatterAggregate,
    FormatterYamlObject,
    HandlerAggregate,
    HandlerYamlObject,
    LoggerAggregate,
    LoggerYamlObject,
    TransferObject
)
from ..utils import Yaml

# *** repos

# ** repo: logging_yaml_repository
class LoggingYamlRepository(LoggingService):
    '''
    YAML-backed repository for logging configurations (formatters, handlers, loggers).
    '''

    # * attribute: yaml_file
    yaml_file: str

    # * attribute: default_role
    default_role: str

    # * attribute: encoding
    encoding: str

    # * init
    def __init__(self, logging_yaml_file: str, encoding: str = 'utf-8'):
        '''
        Initialize the logging YAML repository.

        :param logging_yaml_file: Path to YAML logging config file
        :type logging_yaml_file: str
        :param encoding: File encoding (default 'utf-8')
        :type encoding: str
        '''

        # Set the repository attributes.
        self.yaml_file = logging_yaml_file
        self.default_role = 'to_data.yaml'
        self.encoding = encoding

    # * method: list_all
    def list_all(self) -> Tuple[List[FormatterAggregate], List[HandlerAggregate], List[LoggerAggregate]]:
        '''
        List all formatter, handler, and logger configurations.

        :return: Tuple of (formatters, handlers, loggers)
        :rtype: Tuple[List[FormatterAggregate], List[HandlerAggregate], List[LoggerAggregate]]
        '''

        # Load the logging settings data from the yaml configuration file.
        data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load(
            data_factory=lambda d: LoggingSettingsYamlObject.from_data(**d),
            start_node=lambda d: d.get('logging', {})
        )

        # Return the formatters, handlers, and loggers.
        return (
            [f.map() for f in data.formatters.values()],
            [h.map() for h in data.handlers.values()],
            [l.map() for l in data.loggers.values()]
        )

    # * method: save_formatter
    def save_formatter(self, formatter: FormatterAggregate):
        '''
        Save/update a formatter configuration.

        :param formatter: The formatter configuration to save.
        :type formatter: FormatterAggregate
        '''

        # Create formatter data object from the model.
        formatter_data = TransferObject.from_model(
            FormatterYamlObject,
            formatter
        )

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load()

        # Update the formatter entry.
        full_data.setdefault('logging', {}).setdefault('formatters', {})[formatter.id] = formatter_data.to_primitive(self.default_role)

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding
        ).save(data=full_data)

    # * method: save_handler
    def save_handler(self, handler: HandlerAggregate):
        '''
        Save/update a handler configuration.

        :param handler: The handler configuration to save.
        :type handler: HandlerAggregate
        '''

        # Create handler data object from the model.
        handler_data = TransferObject.from_model(
            HandlerYamlObject,
            handler
        )

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load()

        # Update the handler entry.
        full_data.setdefault('logging', {}).setdefault('handlers', {})[handler.id] = handler_data.to_primitive(self.default_role)

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding
        ).save(data=full_data)

    # * method: save_logger
    def save_logger(self, logger: LoggerAggregate):
        '''
        Save/update a logger configuration.

        :param logger: The logger configuration to save.
        :type logger: LoggerAggregate
        '''

        # Create logger data object from the model.
        logger_data = TransferObject.from_model(
            LoggerYamlObject,
            logger
        )

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load()

        # Update the logger entry.
        full_data.setdefault('logging', {}).setdefault('loggers', {})[logger.id] = logger_data.to_primitive(self.default_role)

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding
        ).save(data=full_data)

    # * method: delete_formatter
    def delete_formatter(self, formatter_id: str):
        '''
        Delete a formatter by ID (idempotent).

        :param formatter_id: The ID of the formatter to delete.
        :type formatter_id: str
        '''

        # Load all formatters data from the yaml file.
        formatter_data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load(
            start_node=lambda d: d.get('logging', {}).get('formatters', {})
        )

        # Pop the formatter data whether it exists or not.
        formatter_data.pop(formatter_id, None)

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load()

        # Update the formatters section.
        full_data.setdefault('logging', {})['formatters'] = formatter_data

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding
        ).save(data=full_data)

    # * method: delete_handler
    def delete_handler(self, handler_id: str):
        '''
        Delete a handler by ID (idempotent).

        :param handler_id: The ID of the handler to delete.
        :type handler_id: str
        '''

        # Load all handlers data from the yaml file.
        handler_data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load(
            start_node=lambda d: d.get('logging', {}).get('handlers', {})
        )

        # Pop the handler data whether it exists or not.
        handler_data.pop(handler_id, None)

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load()

        # Update the handlers section.
        full_data.setdefault('logging', {})['handlers'] = handler_data

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding
        ).save(data=full_data)

    # * method: delete_logger
    def delete_logger(self, logger_id: str):
        '''
        Delete a logger by ID (idempotent).

        :param logger_id: The ID of the logger to delete.
        :type logger_id: str
        '''

        # Load all loggers data from the yaml file.
        logger_data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load(
            start_node=lambda d: d.get('logging', {}).get('loggers', {})
        )

        # Pop the logger data whether it exists or not.
        logger_data.pop(logger_id, None)

        # Load the full configuration file.
        full_data = Yaml(
            self.yaml_file,
            encoding=self.encoding
        ).load()

        # Update the loggers section.
        full_data.setdefault('logging', {})['loggers'] = logger_data

        # Persist the updated configuration file.
        Yaml(
            self.yaml_file,
            mode='w',
            encoding=self.encoding
        ).save(data=full_data)



"""Tiferet Utils Exports"""

# *** exports

# ** app
from .file import FileLoader, FileLoader as File
from .json import JsonLoader, JsonLoader as Json
from .yaml import YamlLoader, YamlLoader as Yaml
from .csv import CsvLoader, CsvLoader as Csv, CsvDictLoader, CsvDictLoader as CsvDict
from .sqlite import SqliteClient, SqliteClient as Sqlite



"""Tiferet Domain Feature"""

# *** imports

# ** app
from .settings import (
    DomainObject,
    StringType,
    BooleanType,
    ListType,
    DictType,
    ModelType,
)

# *** models

# ** model: feature_step
class FeatureStep(DomainObject):
    '''
    A base step in a feature workflow.
    '''

    # * attribute: type
    type = StringType(
        choices=['event'],
        default='event',
        metadata=dict(
            description='The type of the feature step.'
        )
    )

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the feature step.'
        )
    )


# ** model: feature_event
class FeatureEvent(FeatureStep):
    '''
    A feature event step that executes a domain event from the container.
    '''

    # * attribute: service_id
    service_id = StringType(
        required=True,
        metadata=dict(
            description='The service configuration ID for the feature event.'
        )
    )

    # * attribute: flags
    flags = ListType(
        StringType(),
        default=[],
        metadata=dict(
            description='List of feature flags that activate this event.'
        )
    )

    # * attribute: parameters
    parameters = DictType(
        StringType(),
        default={},
        metadata=dict(
            description='The custom parameters for the feature event.'
        )
    )

    # * attribute: return_to_data (obsolete)
    return_to_data = BooleanType(
        default=False,
        metadata=dict(
            description='Whether to return the feature event result to the feature data context.'
        )
    )

    # * attribute: data_key
    data_key = StringType(
        metadata=dict(
            description='The data key to store the feature event result in if Return to Data is True.'
        )
    )

    # * attribute: pass_on_error
    pass_on_error = BooleanType(
        default=False,
        metadata=dict(
            description='Whether to pass on the error if the feature event fails.'
        )
    )


# ** model: feature
class Feature(DomainObject):
    '''
    A feature object.
    '''

    # * attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The unique identifier of the feature.'
        )
    )

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the feature.'
        )
    )

    # * attribute: flags
    flags = ListType(
        StringType(),
        default=[],
        metadata=dict(
            description='List of feature flags that activate this entire feature.'
        )
    )

    # * attribute: description
    description = StringType(
        metadata=dict(
            description='The description of the feature.'
        )
    )

    # * attribute: group_id
    group_id = StringType(
        required=True,
        metadata=dict(
            description='The context group identifier for the feature.'
        )
    )

    # * attribute: feature_key
    feature_key = StringType(
        required=True,
        metadata=dict(
            description='The key of the feature.'
        )
    )

    # * attribute: steps
    steps = ListType(
        ModelType(FeatureEvent),
        default=[],
        metadata=dict(
            description='The step workflow for the feature.'
        )
    )

    # * attribute: log_params
    log_params = DictType(
        StringType(),
        default={},
        metadata=dict(
            description='The parameters to log for the feature.'
        )
    )

    # * method: get_step
    def get_step(self, position: int) -> FeatureStep | None:
        '''
        Get the feature step at the given position, or None if the
        index is out of range or invalid.

        :param position: The index of the step to retrieve.
        :type position: int
        :return: The FeatureStep at the position, or None.
        :rtype: FeatureStep | None
        '''

        # Attempt to retrieve the step at the specified index, returning
        # None if the index is out of range or invalid.
        try:
            return self.steps[position]
        except (IndexError, TypeError):
            return None



"""Tiferet Exceptions (Assets)"""

# *** imports

# ** core
from typing import Dict, Any
import json

# *** exceptions

# ** exception: tiferet_error
class TiferetError(Exception):
    '''
    The TiferetError is the base exception for all Tiferet-related errors.
    It extends the built-in Exception class.
    '''
    
    # * attribute: error_code
    error_code: str

    # * attribute
    kwargs: Dict[str, Any]

    # * init
    def __init__(self, error_code: str, message: str = None, **kwargs):
        '''
        Initialize the TiferetError with an error code, message, and additional arguments.

        :param error_code: The error code.
        :type error_code: str
        :param message: The error message.
        :type message: str
        :param kwargs: Additional error keyword arguments.
        :type kwargs: dict
        '''

        # Set the error code and additional arguments.
        self.error_code = error_code
        self.kwargs = kwargs

        # Initialize base exception with error data.
        super().__init__(
            json.dumps({
                'error_code': error_code,
                'message': message,
                **kwargs
            })
        )

# ** exception: tiferet_api_error
class TiferetAPIError(TiferetError):
    '''
    The TiferetAPIError is the exception returned for all Tiferet API-related errors by default.
    '''
    
    # * attribute: name
    name: str

    # * attribute: message
    message: str

    # * init
    def __init__(self, error_code: str, name: str,  message: str, **kwargs):
        '''
        Initialize the TiferetError with an error code, message, and additional arguments.

        :param error_code: The error code.
        :type error_code: str
        :param name: A descriptive name for the error.
        :type name: str
        :param message: The error message.
        :type message: str
        :param kwargs: Additional error keyword arguments.
        :type kwargs: dict
        '''

        # Set the name and the message.
        self.name = name
        self.message = message

        # Initialize base exception with error data.
        super().__init__(
            error_code=error_code,
            message=message,
            **kwargs
        )
